# coding=utf-8
"""
Implementation for:
- TaggedLabelFrame: a custom LabelFrame widget that allows child widgets to be identified by tags
"""
import tkinter as tk
from tkinter import ttk

from UEVaultManager.tkgui.modules.ExtendedWidgetClasses import WidgetType, ExtendedWidget, ExtendedEntry, ExtendedText, ExtendedLabel, \
    ExtendedCheckButton
from UEVaultManager.tkgui.modules.functions import log_error, log_warning, log_debug


class TaggedLabelFrame(ttk.LabelFrame):
    """
    A custom LabelFrame widget that allows child widgets to be identified by tags.
    :param args: Args to pass to the widget
    :param kwargs: Kwargs to pass to the widget
    :return: TaggedLabelFrame instance
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tagged_child = {}
        self.pack_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.X, 'expand': False}
        self.pack_fw_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.X, 'expand': True}  # full width
        self.lblf_fw_options = {'ipadx': 2, 'ipady': 2, 'fill': tk.X, 'expand': True}

    def add_child(
        self,
        tag: str,
        widget_type: WidgetType.ENTRY,
        default_content=None,
        width=None,
        height=None,
        label=None,
        layout_option='',
        images_folder=None,
        add_label_before=True,
        focus_out_callback=None,
        click_on_callback=None
    ) -> None:
        """
        Adds a child widget to the LabelFrame and associates it with the given tag.
        Note: we can not use command parameter to manage callback here because it should be transmited
        to the parent widget and in that case tag won't be available as an indentificator
        :param tag: Tag to search for (case-insensitive)
        :param widget_type: Type of widget to add. A string value of 'text', 'checkbutton', or 'entry'
        :param width: Width of the child widget. Only used for text widgets
        :param height: Height of the child widget. Only used for text widgets
        :param label: Text to display in the child widget.
        :param default_content: Default content of the child widget
        :param layout_option: Layout options to use. Default, full width
        :param images_folder: folder for image used by some widgets
        :param add_label_before: If True, adds a label before the child widget
        :param focus_out_callback: Callback to call when the child widget loses focus
        :param click_on_callback: Callback to call when the child widget is clicked or checked
        """
        tag = tag.lower()
        frame = ttk.Frame(self)
        frame.pack(**self.lblf_fw_options)
        if add_label_before:
            lbl_name = ttk.Label(frame, text=ExtendedWidget.tag_to_label(tag))
            lbl_name.pack(side=tk.LEFT, **self.pack_options)
        if widget_type == WidgetType.ENTRY:
            child = ExtendedEntry(master=frame, tag=tag, default_content=default_content, height=height, width=width)
        elif widget_type == WidgetType.TEXT:
            child = ExtendedText(master=frame, tag=tag, default_content=default_content, wrap=tk.WORD, height=height, width=width)
        elif widget_type == WidgetType.LABEL:
            child = ExtendedLabel(master=frame, tag=tag, default_content=default_content)
        elif widget_type == WidgetType.CHECKBUTTON:
            child = ExtendedCheckButton(master=frame, tag=tag, default_content=default_content, label=label, images_folder=images_folder)
        else:
            error = f'Invalid widget type: {widget_type}'
            log_error(error)
            return

        self._tagged_child[tag] = child

        layout_option = self.pack_fw_options if layout_option == '' else layout_option
        child.pack(side=tk.LEFT, **layout_option)

        if focus_out_callback is not None:
            child.bind('<FocusOut>', lambda event: focus_out_callback(event=event, tag=tag))
        if click_on_callback is not None:
            child.bind('<Button-1>', lambda event: click_on_callback(event=event, tag=tag))

    def get_child_by_tag(self, tag: str):
        """
        Returns the child widget associated with the given tag.
        :param tag: Tag to search for (case-insensitive)
        :return: Child widget
        """
        tag = tag.lower()
        return self._tagged_child.get(tag)

    def get_children(self) -> dict:
        """
        Returns the dictionary of tagged children.
        :return: A dictionary of tagged children
        """
        return self._tagged_child

    def set_default_content(self, tag='') -> None:
        """
        Sets the default content of the child widget associated with the given tag.
        :param tag: Tag to search for (case-insensitive)
        """
        tag = tag.lower()
        widget = self.get_child_by_tag(tag)
        if widget is not None:
            widget.set_content(widget.default_content)

    def set_child_values(self, tag='', content='', label=None, row=-1, col=-1) -> None:
        """
        Sets the content of the child widget associated with the given tag.
        Also sets its row and column index.
        :param tag: Tag to search for (case-insensitive)
        :param label: Text to set (only for CheckButton widget)
        :param content: Content to set
        :param row: Row index to set
        :param col: Column index to set
        """
        tag = tag.lower()
        widget = self.get_child_by_tag(tag)
        if widget is not None:
            widget.set_content(content)
            widget.row = row
            widget.col = col
            if label is not None:
                try:
                    widget.set_label(label)
                except AttributeError:
                    log_debug(f'Widget with tag {tag} does not have a set_label method')
        else:
            log_warning(f'Widget with tag {tag} not found')



# coding: utf-8
"""
Implementation for:
- InvalidCredentialsError: Exception raised when the credentials are invalid
"""


class InvalidCredentialsError(Exception):
    """
    Exception raised when the credentials are invalid
    """
    pass



# coding=utf-8
"""
global variables and references to global objects
"""
# circular import error
# import UEVaultManager.tkgui.modules.DisplayContentWindowClass as DisplayContentWindow
import UEVaultManager.tkgui.modules.EditCellWindowClass as EditCellWindow
import UEVaultManager.tkgui.modules.EditRowWindowClass as EditRowWindow
from UEVaultManager.tkgui.modules.GUISettingsClass import GUISettings
from UEVaultManager.tkgui.modules.ProgressWindowsClass import ProgressWindow
from UEVaultManager.tkgui.modules.SaferDictClass import SaferDict

# references to global objects
edit_cell_window_ref: EditCellWindow = None
edit_row_window_ref: EditRowWindow = None
# circular import error
# display_content_window_ref: DisplayContentWindow = None
display_content_window_ref = None
# noinspection PyTypeChecker
progress_window_ref: ProgressWindow = None
# reference to the cli object of the UEVM main app (the main one, it gives all access to all the features)
# if empty, direct access to its features from this script won't be available and a message will be displayed instead
# noinspection PyTypeChecker
UEVM_cli_ref = None  # avoid importing classes from the UEVM main app here because it can cause circular dependencies when importing the module
# noinspection PyTypeChecker
UEVM_gui_ref = None  # avoid importing classes from the UEVM GUI class here because it can cause circular dependencies when importing the module
#  reference to the log object of the UEVM main app.
#  If empty, log will be message printed in the console
UEVM_log_ref = None
#  reference to the default command line parser (used for help button in gui).
UEVM_parser_ref = None

# global variables
# noinspection PyTypeChecker
UEVM_cli_args: SaferDict = {}
UEVM_filter_category = ''

s = GUISettings()  # using the shortest variable name for GUISettings for convenience


# options that can be changed in the GUI
def set_args_force_refresh(value: bool) -> None:
    """
    Set the value of the argument force_refresh. Mandadory fot the associated ttk.ckbutton to work
    :param value: True or False
    """
    UEVM_cli_args['force_refresh'] = value


def set_args_debug(value: bool) -> None:
    """
    Set the value of the argument debug. Mandadory fot the associated ttk.ckbutton to work
    :param value: True or False
    """
    UEVM_cli_args['debug'] = value


def set_args_offline(value: bool) -> None:
    """
    Set the value of the argument offline. Mandadory fot the associated ttk.ckbutton to work
    :param value:  True or False
    """
    UEVM_cli_args['offline'] = value



# coding=utf-8
"""
Implementation for:
- EditRowWindow: the window to edit a row
"""
import tkinter as tk
from tkinter import ttk

from ttkbootstrap.constants import *

import UEVaultManager.tkgui.modules.functions as gui_f  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.functions_no_deps as gui_fn  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience


class EditRowWindow(tk.Toplevel):
    """
    The window to edit a row
    :param parent: the parent window
    :param title: the title of the window
    :param width: the width of the window
    :param height: the height of the window
    :param icon: the icon of the window
    :param screen_index: the index of the screen on which the window will be displayed
    :param editable_table: the table to edit
    """

    def __init__(self, parent, title: str, width: int = 600, height: int = 800, icon=None, screen_index: int = 0, editable_table=None):
        super().__init__(parent)
        self.title(title)
        style = gui_fn.set_custom_style(gui_g.s.theme_name, gui_g.s.theme_font)
        self.style = style
        geometry = gui_fn.center_window_on_screen(screen_index, height, width)
        self.geometry(geometry)
        gui_fn.set_icon_and_minmax(self, icon)
        self.resizable(True, False)

        self.editable_table = editable_table
        self.must_save = False
        self.initial_values = []
        self.width = width
        # the photoimage is stored is the variable to avoid garbage collection
        # see: https://stackoverflow.com/questions/30210618/image-not-getting-displayed-on-tkinter-through-label-widget
        self.canvas_image = None

        self.control_frame = self.ControlFrame(self)
        self.content_frame = self.ContentFrame(self)

        self.control_frame.pack(ipadx=5, ipady=5, fill=tk.X)
        self.content_frame.pack(ipadx=5, ipady=5, padx=5, pady=5, fill=tk.X)

        self.bind('<Key>', self.on_key_press)
        self.protocol("WM_DELETE_WINDOW", self.on_close)

        gui_g.edit_row_window_ref = self

    class ContentFrame(ttk.Frame):
        """
        The frame containing the editable fields
        :param container: the parent window
        """

        def __init__(self, container):
            super().__init__(container)

    class ControlFrame(ttk.Frame):
        """
        The frame containing the buttons
        :param container: the parent window
        """

        def __init__(self, container):
            super().__init__(container)
            pack_def_options = {'ipadx': 2, 'ipady': 2, 'fill': tk.X, 'anchor': tk.NW}
            grid_def_options = {'ipadx': 5, 'ipady': 5, 'padx': 2, 'pady': 2, 'sticky': tk.NSEW}

            lblf_navigation = ttk.LabelFrame(self, text='Navigation')
            lblf_navigation.grid(row=0, column=0, **grid_def_options)
            btn_prev = ttk.Button(lblf_navigation, text='Prev Asset', command=container.prev_asset)
            btn_prev.pack(**pack_def_options, side=tk.LEFT)
            btn_next = ttk.Button(lblf_navigation, text='Next Asset', command=container.next_asset)
            btn_next.pack(**pack_def_options, side=tk.RIGHT)

            lbf_preview = ttk.LabelFrame(self, text="Image Preview")
            lbf_preview.grid(row=0, column=1, **grid_def_options)
            canvas_image = tk.Canvas(lbf_preview, width=gui_g.s.preview_max_width, height=gui_g.s.preview_max_height, highlightthickness=0)
            canvas_image.pack()
            canvas_image.create_rectangle((0, 0), (gui_g.s.preview_max_width, gui_g.s.preview_max_height), fill='black')

            lblf_actions = ttk.LabelFrame(self, text='Actions')
            lblf_actions.grid(row=0, column=2, **grid_def_options)
            btn_open_url = ttk.Button(lblf_actions, text="Open URL", command=container.open_asset_url)
            btn_open_url.pack(**pack_def_options, side=tk.LEFT)
            # noinspection PyArgumentList
            btn_on_close = ttk.Button(lblf_actions, text='Close', command=container.on_close, bootstyle=WARNING)
            btn_on_close.pack(**pack_def_options, side=tk.RIGHT)
            # noinspection PyArgumentList
            bnt_save = ttk.Button(lblf_actions, text='Save Changes', command=container.save_change, bootstyle=INFO)
            bnt_save.pack(**pack_def_options, side=tk.RIGHT)

            # Configure the columns to take all the available width
            self.columnconfigure(0, weight=1)
            self.columnconfigure(1, weight=2)
            self.columnconfigure(2, weight=1)

            self.canvas_image = canvas_image

    def on_close(self, _event=None) -> None:
        """
        Event when the window is closing
        :param _event: the event that triggered the call of this function
        """
        current_values = self.editable_table.get_edited_row_values()
        # current_values is empty is save_button has been pressed because global variables have been cleared in save_changes()
        self.must_save = current_values and self.initial_values != current_values
        if self.must_save:
            if gui_f.box_yesno('Changes have been made in the window. Do you want to keep them ?'):
                self.save_change()
        self.close_window()

    def on_key_press(self, event) -> None:
        """
        Event when a key is pressed
        :param event: the event that triggered the call of this function
        """
        if event.keysym == 'Escape':
            self.on_close()
        elif event.keysym == 'Return':
            self.save_change()

    def close_window(self) -> None:
        """
        Close the window
        """
        gui_g.edit_row_window_ref = None
        self.editable_table.reset_style()
        self.destroy()

    def save_change(self) -> None:
        """
        Save the changes (Wrapper)
        """
        self.must_save = False
        self.editable_table.save_edit_row_record()

    def prev_asset(self) -> None:
        """
        Go to the previous asset (Wrapper)
        """
        self.editable_table.move_to_prev_record()

    def next_asset(self) -> None:
        """
        Go to the next asset (Wrapper)
        """
        self.editable_table.move_to_next_record()

    def open_asset_url(self) -> None:
        """
        Open the asset URL (Wrapper)
        """
        self.editable_table.open_asset_url()



# coding=utf-8
"""
UEVaultManager setup file.
"""
import datetime

year = datetime.date.today().year

__name__ = 'UEVaultManager'
__version__ = '1.6.4'
# 0 Pegasus Seiya
# 1 Dragon Shiryu
# 2 Cygnus Hyoga
# 3 Andromeda Shun
# 4 Phoenix Ikki
# 5 Leo Aiolia
# 5 Virgo Shaka
__codename__ = 'Virgo+4'
# 6 Libra Dohko
# 7 Scorpio Milo
# 8 Sagittarius Aiolos
# 9 Capricorn Shura
# 10 Aquarius Camus
# 11 Pisces Aphrodite
__author__ = 'Laurent Ongaro'
__author_email__ = 'laurent@gameamea.com'
__description__ = 'Free and open-source replacement for the Epic Games Launcher application, mainly to manage the assets for Unreal Engine'
__copyright__ = f'{year} {__author__}'
__license__ = 'GPL-3'
__url__ = 'https://github.com/LaurentOngaro/UEVaultManager'



# coding=utf-8
"""
Utilities functions and tools
These functions depend on the globals.py module and can generate circular dependencies when imported
"""
import datetime
import logging
import os
import sys
import time
import tkinter as tk
from io import BytesIO
from tkinter import messagebox

import requests
from PIL import ImageTk, Image
from termcolor import colored

from UEVaultManager.tkgui.modules import globals as gui_g


def log_format_message(name: str, levelname: str, message: str) -> str:
    """
    Format the log message
    :param name: the name of the logger
    :param levelname: the level of log
    :param message: the message to format
    :return: the formatted message
    """
    return f'[{name}] {levelname}: {message}'


def box_message(msg: str, level='info'):
    """
    Display a message box with the given message
    :param msg: the message to display
    :param level: the level of the message (info, warning, error)
    """
    level = level.lower()
    if level == 'warning':
        log_warning(msg)
        messagebox.showwarning(title=gui_g.s.app_title, message=msg)
    elif level == 'error':
        messagebox.showerror(title=gui_g.s.app_title, message=msg)
        log_error(msg)
        # done in log_error
        # exit(1)
    else:
        log_info(msg)
        messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def box_yesno(msg: str) -> bool:
    """
    Display a YES/NO message box with the given message
    :param msg: the message to display 
    :return:  True if the user clicked on Yes, False otherwise
    """
    return messagebox.askyesno(title=gui_g.s.app_title, message=msg)


def box_okcancel(msg: str) -> bool:
    """
    Display an OK/CANCEL message box with the given message
    :param msg: the message to display
    :return:  True if the user clicked on Yes, False otherwise
    """
    return messagebox.askokcancel(title=gui_g.s.app_title, message=msg)


def todo_message() -> None:
    """
    Display a message box with a message saying that the feature is not implemented yet
    """
    msg = 'Not implemented yet'
    msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'yellow'))
    print(msg)
    messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def from_cli_only_message() -> None:
    """
    Display a message box with a message saying that the feature is only accessible when running the app using the UEVM cli command options
    """
    msg = 'This feature is only accessible when running these app using the UEVM cli command options. Once the UEVaultManager package installed, Type UEVaultManager -h for more help'
    msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'yellow'))
    print(msg)
    messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def log_info(msg: str) -> None:
    """
    Log an info message
    :param msg: the message to log
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.info(msg)
    else:
        msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'blue'))
        print(msg)


def log_debug(msg: str) -> None:
    """
    Log a debug message. Note that this message will only be logged if the debug mode is enabled
    :param msg: the message to log
    """
    if not gui_g.s.debug_mode:
        return
    if gui_g.UEVM_log_ref is not None:
        # ensure that the debug messages will be logged even if the log level not set to DEBUG in the cli
        log_level = gui_g.UEVM_log_ref.level
        if log_level == logging.DEBUG:
            gui_g.UEVM_log_ref.debug(msg)
        else:
            gui_g.UEVM_log_ref.info(msg)
    else:
        msg = log_format_message(gui_g.s.app_title, 'Debug', colored(msg, 'light_grey'))
        print(msg)


def log_warning(msg: str) -> None:
    """
    Log a warning message
    :param msg: the message to log
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.info(msg)
    else:
        msg = log_format_message(gui_g.s.app_title, 'Warning', colored(msg, 'orange'))
        print(msg)


def log_error(msg: str) -> None:
    """
    Log an error message
    :param msg: the message to log. Note that the app will exit after logging the message
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.error(msg)
    else:
        msg = log_format_message(gui_g.s.app_title, 'Error', colored(msg, 'red', 'bold'))
        print(msg)
    if gui_g.UEVM_gui_ref is not None:
        gui_g.UEVM_gui_ref.quit()
    sys.exit(1)


def convert_to_datetime(value: str) -> datetime.datetime:
    """
    Convert a value to a datetime
    :param value: the value to convert. If the value is not a datetime, it will be converted to a string and then to a datetime.
    :return: the datetime value
    """
    try:
        return datetime.datetime.strptime(value, gui_g.s.csv_datetime_format)
    except ValueError:
        return datetime.datetime.today()


# noinspection DuplicatedCode
def resize_and_show_image(image: Image, canvas: tk.Canvas) -> None:
    """
    Resize the given image and display it in the given canvas
    :param image: the image to display
    :param canvas: the canvas to display the image in
    """
    # Resize the image while keeping the aspect ratio
    target_height = gui_g.s.preview_max_height
    aspect_ratio = float(image.width) / float(image.height)
    target_width = int(target_height * aspect_ratio)
    resized_image = image.resize((target_width, target_height), Image.ANTIALIAS)
    tk_image = ImageTk.PhotoImage(resized_image)
    # Calculate center coordinates
    x = (canvas.winfo_width() - tk_image.width()) // 2
    y = (canvas.winfo_height() - tk_image.height()) // 2
    canvas.create_image(x, y, anchor=tk.NW, image=tk_image)
    canvas.image = tk_image


def show_asset_image(image_url: str, canvas_image=None) -> None:
    """
    Show the image of the given asset in the given canvas
    :param image_url: the url of the image to display
    :param canvas_image: the canvas to display the image in
    """
    if canvas_image is None or image_url == '' or str(image_url).lower() == gui_g.s.empty_cell:
        return
    try:
        # noinspection DuplicatedCode
        if not os.path.isdir(gui_g.s.cache_folder):
            os.mkdir(gui_g.s.cache_folder)
        image_filename = os.path.join(gui_g.s.cache_folder, os.path.basename(image_url))
        # Check if the image is already cached
        if os.path.isfile(image_filename) and (time.time() - os.path.getmtime(image_filename)) < gui_g.s.image_cache_max_time:
            # Load the image from the cache folder
            image = Image.open(image_filename)
        else:
            response = requests.get(image_url)
            image = Image.open(BytesIO(response.content))

            with open(image_filename, "wb") as f:
                f.write(response.content)
        resize_and_show_image(image, canvas_image)
    except Exception as error:
        log_warning(f"Error showing image: {error}")


def show_default_image(canvas_image=None) -> None:
    """
    Show the default image in the given canvas
    :param canvas_image: the canvas to display the image in
    """
    if canvas_image is None:
        return
    try:
        # Load the default image
        if os.path.isfile(gui_g.s.default_image_filename):
            def_image = Image.open(gui_g.s.default_image_filename)
            # noinspection PyTypeChecker
            resize_and_show_image(def_image, canvas_image)
    except Exception as error:
        log_warning(f"Error showing default image {gui_g.s.default_image_filename} cwd:{os.getcwd()}: {error}")


def json_print_key_val(json_obj, indent=4, print_result=True, output_on_gui=False) -> None:
    """
    Pretty prints a JSON object in a simple 'key: value' format.
    :param json_obj:  The JSON object to print.
    :param indent: The number of spaces to indent each level.
    :param print_result: Determines whether to print the result.
    :param output_on_gui: Determines whether to print the result on the GUI.
    :return: The pretty printed JSON object.
    """

    def _process(obj, level=0):
        lines = []
        indent_str = ' ' * indent * level

        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, (dict, list)):
                    lines.append(f'{indent_str}{key}:')
                    lines.extend(_process(value, level + 1))
                else:
                    lines.append(f'{indent_str}{key}: {value}')
        elif isinstance(obj, list):
            for item in obj:
                lines.extend(_process(item, level))
        else:
            lines.append(f'{indent_str}{obj}')

        return lines

    result = '\n'.join(_process(json_obj))

    if print_result:
        if output_on_gui and gui_g.display_content_window_ref is not None:
            gui_g.display_content_window_ref.display(result)
        else:
            print(result)


def custom_print(text='', keep_mode=True) -> None:
    """
    Print the given text on the GUI if it's available, otherwise print it on the console²
    :param text: the text to print
    :param keep_mode: whether to keep the existing content when adding a new one
    """
    if gui_g.display_content_window_ref is not None:
        gui_g.display_content_window_ref.display(content=text, keep_mode=keep_mode)
    else:
        print(text)



# coding=utf-8
"""
Implementation for:
- WebImage: class to download an image from an url and get it as a PhotoImage
"""
from io import BytesIO

import requests
from PIL import ImageTk, Image

from UEVaultManager.tkgui.modules.functions import log_warning


class WebImage:
    """
    Class to download an image from an url and get it as a PhotoImage
    :param url: the url of the image to download
    """

    def __init__(self, url: str = None):
        # if no URL is given, return
        if url is None or url == '':
            return
        # initialize instance variables
        self.__image_pil = None
        self.__image_tk = None
        self.url = url
        try:
            # use requests to get the image content as bytes
            response = requests.get(url)
            # create a PIL.Image object from the bytes
            self.__image_pil = Image.open(BytesIO(response.content))
            # create a PhotoImage object from the PIL.Image
            self.__image_tk = ImageTk.PhotoImage(self.__image_pil)
        except Exception as error:
            # log a warning if image cannot be downloaded or opened
            log_warning(f'image could not be read from url {self.url}.\nError:{error}')

    def get(self) -> ImageTk.PhotoImage:
        """
        Get the downloaded image
        :return: the image
        """
        return self.__image_tk

    def get_resized(self, new_width: int, new_height: int) -> ImageTk.PhotoImage:
        """
        Get the downloaded image resized to the given size
        :param new_width: width of the resized image
        :param new_height: height of the resized image
        :return: the resized image
        """
        try:
            # resize the PIL.Image object in place
            self.__image_pil.thumbnail((new_width, new_height))
            # create a PhotoImage object from the resized PIL.Image
            self.__image_tk = ImageTk.PhotoImage(self.__image_pil)
        except Exception as error:
            # log a warning if image cannot be resized
            log_warning(f'Could not get resized image from url {self.url}.\nError:{error}')
        return self.__image_tk



# coding=utf-8
"""
Utilities functions and tools
These functions DO NOT depend on the globals.py module and be freely imported
"""
import ctypes as ct
import os
import sys

import ttkbootstrap as ttk
from screeninfo import get_monitors


def path_from_relative_to_absolute(path: str) -> str:
    """
    Build the path of the file to reference relative to the currently running script
    :param path: the relative path to the file. If the path is already absolute, it is returned as is
    :return: the absolute path of the file
    """

    if os.path.isabs(path):
        return path

    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        # noinspection PyProtectedMember,PyUnresolvedReferences
        # base_path = sys._MEIPASS
        current_script_directory = sys._MEIPASS
        # when launched from pyinstaller, the path has been copied at the root of the folder where the script is run. So we nne to remove all the relative path indicator
        path = os.path.basename(path)
    except AttributeError:
        # base_path = os.path.abspath(".")
        current_script_path = os.path.abspath(__file__)
        current_script_directory = os.path.dirname(current_script_path)

    absolute_path = os.path.join(current_script_directory, path)
    absolute_path = os.path.abspath(absolute_path)
    # messagebox.showinfo('info', 'absolute_path: ' + absolute_path)
    return absolute_path


def center_window_on_screen(screen_index: int, height: int, width: int) -> str:
    """
    Calculate the geometry of the window to display in the center of the given screen
    :param screen_index: the index of the screen to use
    :param height: the height of the window
    :param width: the width of the window
    :return: the geometry string to use to display the window in the center of the screen
    """
    monitors = get_monitors()
    if screen_index > len(monitors):
        print(f'The screen #{screen_index} is not available. Using 0 as screen index.')  # no use of log functions here to prevent circular import
        screen_index = 0
    # Position the window in the center of the screen
    target_screen = monitors[screen_index]
    screen_width = target_screen.width
    screen_height = target_screen.height
    x = target_screen.x + (screen_width-width) // 2
    y = target_screen.y + (screen_height-height) // 2
    geometry: str = f'{width}x{height}+{x}+{y}'
    return geometry


def set_custom_style(theme_name='lumen', font=('Arial', 10, 'normal')):
    """
    Set the custom style for the application
    :return: the style object
    """
    style = ttk.Style(theme_name)
    # option possible for ttk widgets:
    # TButton, TCheckbutton, TCombobox, TEntry, TFrame, TLabel, TLabelFrame, TMenubutton, TNotebook, TProgressbar, TRadiobutton,
    # TScale, TScrollbar, TSeparator, TSizegrip, Treeview, TPanedwindow,
    # Horizontal.TProgressbar or Vertical.TProgressbar (depending on the orient option),
    # Horizontal.TScale or Vertical.TScale (depending on the orient option),
    # Horizontal.TScrollbar or Vertical.TScrollbar (depending on the orient option)
    style.configure('TLabel', font=font, spacing=1, padding=2)
    style.configure('TButton', font=font, spacing=1, padding=2)
    style.configure('TEntry', font=font, spacing=1, padding=2)
    style.configure('TFrame', font=font, spacing=1, padding=1)
    style.configure('TCombobox', font=font, spacing=1, padding=1)
    style.configure('TLabelFrame', font=font, spacing=1, padding=1)
    return style


def set_toolbar_style(tk_window) -> None:
    """
    Remove the minimize and maximize buttons from a tkinter window.
    This version is compatible with Windows AND Non-windows OS
    # see https://stackoverflow.com/questions/2969870/removing-minimize-maximize-buttons-in-tkinter
    :param tk_window: the tkinter window
    """
    try:
        set_window_pos = ct.windll.user32.SetWindowPos
        set_window_long = ct.windll.user32.SetWindowLongPtrW
        get_window_long = ct.windll.user32.GetWindowLongPtrW
        get_parent = ct.windll.user32.GetParent
    except AttributeError:
        # Non-windows OS
        print('Non-windows OS detected. No need to remove the minimize and maximize buttons from the window.')
        return
    # Identifiers
    gwl_style = -16
    ws_minimizebox = 131072
    ws_maximizebox = 65536
    swp_nozorder = 4
    swp_nomove = 2
    swp_nosize = 1
    swp_framechanged = 32
    hwnd = get_parent(tk_window.winfo_id())
    old_style = get_window_long(hwnd, gwl_style)  # Get the style
    new_style = old_style & ~ws_maximizebox & ~ws_minimizebox  # New style, without max/min buttons
    set_window_long(hwnd, gwl_style, new_style)  # Apply the new style
    set_window_pos(hwnd, 0, 0, 0, 0, 0, swp_nomove | swp_nosize | swp_nozorder | swp_framechanged)  # Updates


def set_icon_and_minmax(tk_window, icon=None) -> None:
    """
    Set the icon and remove the min/max buttons of the window if no icon is provided
    :param tk_window:
    :param icon:
    """
    if icon is None:
        # remove the min/max buttons of the window
        # this code works on Window only
        # tk_window.attributes('-toolwindow', True)
        tk_window.after(300, lambda: set_toolbar_style(tk_window))
    else:
        # windows only (remove the minimize/maximize buttons and the icon)
        icon = path_from_relative_to_absolute(icon)
        if icon != '' and os.path.isfile(icon):
            tk_window.iconbitmap(icon)


def create_empty_file(file_path: str) -> str:
    """
    Create an empty file
    :param file_path: the path of the file to create
    :return: the path of the file
    """
    path = os.path.dirname(file_path)
    file = os.path.basename(file_path)
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except (OSError, PermissionError) as e:
            print(f'Error while creating the directory {path}: {e}')
            if home_dir := os.environ.get('XDG_CONFIG_HOME'):
                path = os.path.join(home_dir, 'UEVaultManager')
            else:
                path = os.path.expanduser('~/.config/UEVaultManager')
            if not os.path.exists(path):
                os.makedirs(path)
            file_path = os.path.normpath(os.path.join(path, file))
            print(f'The following file {file_path} will be used as default')
    open(file_path, 'a').close()
    return file_path


def convert_to_bool(value) -> bool:
    """
    Convert a value to a boolean
    :param value: the value to convert. If the value is not a boolean, it will be converted to a string and then to a boolean.
    :return:
    """
    try:
        if str(value).lower() in ('1', '1.0', 'true', 'yes', 'y', 't'):
            return True
        else:
            return False
    except ValueError:
        return False



# coding=utf-8
"""
Environment detection functions.
"""
import os
import sys


def is_pyinstaller() -> bool:
    """
    Returns True if the program is running in a PyInstaller bundle.
    """
    return getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS')


def is_windows_mac_or_pyi() -> bool:
    """
    Returns True if the program is running on Windows, macOS, or in a PyInstaller bundle.
    """
    return is_pyinstaller() or os.name == 'nt' or sys.platform == 'darwin'



# coding=utf-8
"""
Implementation for:
- AppCore: handles most of the lower level interaction with the downloader, lfs, and api components to make writing CLI/GUI code easier and cleaner and avoid duplication.
"""
import json
import logging
import os
import sys
import time
from base64 import b64decode
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from hashlib import sha1
from locale import getlocale, LC_CTYPE
from platform import system
from threading import current_thread, enumerate as thread_enumerate
from urllib.parse import urlparse

from requests import session
from requests.exceptions import HTTPError, ConnectionError

import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
# noinspection PyPep8Naming
from UEVaultManager import __version__ as UEVM_version
from UEVaultManager.api.egs import EPCAPI, GrabResult
from UEVaultManager.api.uevm import UEVMAPI
from UEVaultManager.lfs.egl import EPCLFS
from UEVaultManager.lfs.uevmlfs import UEVMLFS
from UEVaultManager.models.app import *
from UEVaultManager.models.exceptions import *
from UEVaultManager.models.json_manifest import JSONManifest
from UEVaultManager.models.manifest import Manifest
from UEVaultManager.tkgui.modules.functions import box_message
from UEVaultManager.utils.cli import check_and_create_path
from UEVaultManager.utils.egl_crypt import decrypt_epic_data
from UEVaultManager.utils.env import is_windows_mac_or_pyi

# The heading dict contains the title of each column and a boolean value to know if its contents must be preserved if it already exists in the output file (To Avoid overwriting data changed by the user in the file)

# ToDo: instead of true/false return values for success/failure actually raise an exception that the CLI/GUI
#  can handle to give the user more details. (Not required yet since there's no GUI so log output is fine)


class AppCore:
    """
    AppCore handles most of the lower level interaction with
    the downloader, lfs, and api components to make writing CLI/GUI
    code easier and cleaner and avoid duplication.
    :param override_config: path to a config file to use instead of the default
    :param timeout: timeout in seconds for requests
    """
    _egl_version = '11.0.1-14907503+++Portal+Release-Live'

    def __init__(self, override_config=None, timeout=10.0):
        self.log = logging.getLogger('Core')
        self.egs = EPCAPI(timeout=timeout)
        self.uevmlfs = UEVMLFS(config_file=override_config)
        self.egl = EPCLFS()
        self.uevm_api = UEVMAPI()

        # on non-Windows load the programdata path from config
        if os.name != 'nt':
            self.egl.programdata_path = self.uevmlfs.config.get('UEVaultManager', 'egl_programdata', fallback=None)
            if self.egl.programdata_path and not os.path.exists(self.egl.programdata_path):
                self.log.error(f'Config EGL path ("{self.egl.programdata_path}") is invalid! Disabling sync...')
                self.egl.programdata_path = None
                self.uevmlfs.config.remove_option('UEVaultManager', 'egl_programdata')
                self.uevmlfs.save_config()

        self.local_timezone = datetime.now().astimezone().tzinfo
        self.language_code, self.country_code = ('en', 'US')

        if locale := self.uevmlfs.config.get('UEVaultManager', 'locale', fallback=getlocale(LC_CTYPE)[0]):
            try:
                self.language_code, self.country_code = locale.split('-' if '-' in locale else '_')
                self.log.debug(f'Set locale to {self.language_code}-{self.country_code}')
                # adjust egs api language as well
                self.egs.language_code, self.egs.country_code = self.language_code, self.country_code
            except Exception as error:
                self.log.warning(f'Getting locale failed: {error!r}, falling back to using en-US.')
        elif system() != 'Darwin':  # macOS doesn't have a default locale we can query
            self.log.warning('Could not determine locale, falling back to en-US')

        self.update_available = False
        self.force_show_update = False
        self.webview_killswitch = False
        self.logged_in = False

        self.default_datetime_format = '%y-%m-%d %H:%M:%S'
        # UE assets metadata cache properties
        self.ue_assets_count = 0
        self.cache_is_invalidate = False
        # Delay (in seconds) when UE assets metadata cache will be invalidated. Default value is 15 days
        self.ue_assets_max_cache_duration = 15 * 24 * 3600
        # set to True to add print more information during long operations
        self.verbose_mode = False
        # Create a backup of the output file (when using the --output option) suffixed by a timestamp before creating a new file
        self.create_output_backup = True
        # Set the file name (and path) for logging when an asset is ignored or filtered when running the --list command
        self.ignored_assets_filename_log = ''
        # Set the file name (and path) for logging when an asset is not found on the marketplace when running the --list command
        self.notfound_assets_filename_log = ''
        # Set the file name (and path) for logging when an asset has metadata and extras data are incoherent when running the --list command
        self.bad_data_assets_filename_log = ''
        # Create a backup of the log files that store asset analysis suffixed by a timestamp before creating a new file
        self.create_log_backup = True
        # new file loggers
        self.ignored_logger = None
        self.notfound_logger = None
        self.bad_data_logger = None
        # store time to process metadata and extras update
        self.process_time_average = {'time': 0.0, 'count': 0}
        self.use_threads = False
        self.thread_executor = None
        self.thread_executor_must_stop = False
        self.engine_version_for_obsolete_assets = '4.26'

    def setup_assets_logging(self) -> None:
        """
        Setup logging for ignored, not found and bad data assets
        """

        def create_logger(logger_name: str, filename_log: str) -> None:
            """
            Create a logger for ignored, not found and bad data assets
            :param logger_name:   
            :param filename_log: 
            :return: 
            """
            filename_log = filename_log.replace('~/.config', self.uevmlfs.path)
            if check_and_create_path(filename_log):
                handler = logging.FileHandler(filename_log, mode='w')
                handler.setFormatter(formatter)
                logger = logging.Logger(logger_name, 'INFO')
                logger.addHandler(handler)
                logger.info(message)

        formatter = logging.Formatter('%(message)s')
        message = f"-----\n{datetime.now().strftime(self.default_datetime_format)} Log Started\n-----\n"

        if self.ignored_assets_filename_log:
            create_logger('IgnoredAssets', self.ignored_assets_filename_log)
        if self.notfound_assets_filename_log:
            create_logger('NotFoundAssets', self.notfound_assets_filename_log)
        if self.bad_data_assets_filename_log:
            create_logger('BadDataAssets', self.bad_data_assets_filename_log)

    def auth_sid(self, sid) -> str:
        """
        Handles getting an exchange code from an id
        :param sid: session id
        :return: exchange code
        """
        s = session()
        s.headers.update(
            {
                'X-Epic-Event-Action':
                'login',
                'X-Epic-Event-Category':
                'login',
                'X-Epic-Strategy-Flags':
                '',
                'X-Requested-With':
                'XMLHttpRequest',
                'User-Agent':
                'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                'AppleWebKit/537.36 (KHTML, like Gecko) '
                f'EpicGamesLauncher/{self._egl_version} '
                'UnrealEngine/4.23.0-14907503+++Portal+Release-Live '
                'Chrome/84.0.4147.38 Safari/537.36'
            }
        )
        s.cookies['EPIC_COUNTRY'] = self.country_code.upper()

        # get first set of cookies (EPIC_BEARER_TOKEN etc.)
        _ = s.get('https://www.epicgames.com/id/api/set-sid', params=dict(sid=sid))
        # get XSRF-TOKEN and EPIC_SESSION_AP cookie
        _ = s.get('https://www.epicgames.com/id/api/csrf')
        # finally, get the exchange code
        r = s.post('https://www.epicgames.com/id/api/exchange/generate', headers={'X-XSRF-TOKEN': s.cookies['XSRF-TOKEN']})

        if r.status_code == 200:
            return r.json()['code']

        self.log.error(f'Getting exchange code failed: {r.json()}')
        return ''

    def auth_code(self, code) -> bool:
        """
        Handles authentication via authorization code (either retrieved manually or automatically)
        """
        try:
            self.uevmlfs.userdata = self.egs.start_session(authorization_code=code)
            return True
        except Exception as error:
            self.log.error(f'Logging in failed with {error!r}, please try again.')
            return False

    def auth_ex_token(self, code) -> bool:
        """
        Handles authentication via exchange token (either retrieved manually or automatically)
        """
        try:
            self.uevmlfs.userdata = self.egs.start_session(exchange_token=code)
            return True
        except Exception as error:
            self.log.error(f'Logging in failed with {error!r}, please try again.')
            return False

    def auth_import(self) -> bool:
        """
        Import refresh token from EGL installation and use it for logging in
        :return: True if successful, False otherwise
        :raises ValueError
        """
        self.egl.read_config()
        remember_me_data = self.egl.config.get('RememberMe', 'Data')
        raw_data = b64decode(remember_me_data)
        # data is encrypted
        if raw_data[0] != '{':
            for data_key in self.egl.data_keys:
                try:
                    decrypted_data = decrypt_epic_data(data_key, raw_data)
                    re_data = json.loads(decrypted_data)[0]
                    break
                except Exception as error:
                    self.log.debug(f'Decryption with key {data_key} failed with {error!r}')
            else:
                raise ValueError('Decryption of EPIC launcher user information failed.')
        else:
            re_data = json.loads(raw_data)[0]

        if 'Token' not in re_data:
            raise ValueError('No login session in config')
        refresh_token = re_data['Token']
        try:
            self.uevmlfs.userdata = self.egs.start_session(refresh_token=refresh_token)
            return True
        except Exception as error:
            self.log.error(f'Logging in failed with {error!r}, please try again.')
            return False

    def login(self, force_refresh=False) -> bool:
        """
        Attempts logging in with existing credentials.
        :param force_refresh: if True, force a refresh of the session
        :return: True if successful, False otherwise
        :raises ValueError if no existing credentials or InvalidCredentialsError if the API return an error
        """
        if not self.uevmlfs.userdata:
            raise ValueError('No saved credentials')
        elif self.logged_in and self.uevmlfs.userdata['expires_at']:
            dt_exp = datetime.fromisoformat(self.uevmlfs.userdata['expires_at'][:-1])
            dt_now = datetime.utcnow()
            td = dt_now - dt_exp

            # if session still has at least 10 minutes left we can re-use it.
            if dt_exp > dt_now and abs(td.total_seconds()) > 600:
                return True
            else:
                self.logged_in = False

        # run update check
        if self.update_check_enabled():
            try:
                self.check_for_updates()
            except Exception as error:
                self.log.warning(f'Checking for UEVaultManager updates failed: {error!r}')

        if self.uevmlfs.userdata['expires_at'] and not force_refresh:
            dt_exp = datetime.fromisoformat(self.uevmlfs.userdata['expires_at'][:-1])
            dt_now = datetime.utcnow()
            td = dt_now - dt_exp

            # if session still has at least 10 minutes left we can re-use it.
            if dt_exp > dt_now and abs(td.total_seconds()) > 600:
                self.log.info('Trying to re-use existing login session...')
                try:
                    self.egs.resume_session(self.uevmlfs.userdata)
                    self.logged_in = True
                    return True
                except InvalidCredentialsError as error:
                    self.log.warning(f'Resuming failed due to invalid credentials: {error!r}')
                except Exception as error:
                    self.log.warning(f'Resuming failed for unknown reason: {error!r}')
                # If verify fails just continue the normal authentication process
                self.log.info('Falling back to using refresh token...')

        try:
            self.log.info('Logging in...')
            userdata = self.egs.start_session(self.uevmlfs.userdata['refresh_token'])
        except InvalidCredentialsError:
            self.log.error('Stored credentials are no longer valid! Please login again.')
            self.uevmlfs.invalidate_userdata()
            return False
        except (HTTPError, ConnectionError) as error:
            self.log.error(f'HTTP request for login failed: {error!r}, please try again later.')
            return False

        self.uevmlfs.userdata = userdata
        self.logged_in = True
        return True

    def update_check_enabled(self) -> bool:
        """
        Returns whether update checks are enabled or not
        :return: True if update checks are enabled, False otherwise
        """
        return not self.uevmlfs.config.getboolean('UEVaultManager', 'disable_update_check', fallback=False)

    def update_notice_enabled(self) -> bool:
        """
        Returns whether update notices are enabled or not
        :return: True if update notices are enabled, False otherwise
        """
        if self.force_show_update:
            return True
        return not self.uevmlfs.config.getboolean('UEVaultManager', 'disable_update_notice', fallback=not is_windows_mac_or_pyi())

    def check_for_updates(self, force=False) -> None:
        """
        Checks for updates and sets the update_available flag accordingly
        :param force: force update check
        """

        def version_tuple(v):
            """
            Converts a version string to a tuple of ints
            :param v: version string
            :return:  tuple of ints
            """
            return tuple(map(int, (v.split('.'))))

        cached = self.uevmlfs.get_cached_version()
        version_info = cached['data']
        if force or not version_info or (datetime.now().timestamp() - cached['last_update']) > 24 * 3600:
            version_info = self.uevm_api.get_version_information()
            self.uevmlfs.set_cached_version(version_info)

        web_version = version_info['version']
        self.update_available = version_tuple(web_version) > version_tuple(UEVM_version)

    def get_update_info(self) -> dict:
        """
        Returns update info dict
        :return: update info dict
        """
        return self.uevmlfs.get_cached_version()['data']

    def update_aliases(self, force=False) -> None:
        """
        Updates aliases if enabled
        :param force: force alias update
        """
        _aliases_enabled = not self.uevmlfs.config.getboolean('UEVaultManager', 'disable_auto_aliasing', fallback=False)
        if _aliases_enabled and (force or not self.uevmlfs.aliases):
            self.uevmlfs.generate_aliases()

    def get_assets(self, update_assets=False, platform='Windows') -> List[AppAsset]:
        """
        Returns a list of assets for the given platform.
        :param update_assets: if True, always fetches a new list of assets from the server
        :param platform: platform to fetch assets for
        :return: list of AppAsset objects
        """
        # do not save and always fetch list when platform is overridden
        if not self.uevmlfs.assets or update_assets or platform not in self.uevmlfs.assets:
            # if not logged in, return empty list
            if not self.egs.user:
                return []

            assets = self.uevmlfs.assets.copy() if self.uevmlfs.assets else dict()

            assets.update({platform: [AppAsset.from_egs_json(a) for a in self.egs.get_item_assets(platform=platform)]})

            # only save (and write to disk) if there were changes
            if self.uevmlfs.assets != assets:
                self.uevmlfs.assets = assets

        return self.uevmlfs.assets[platform]

    def get_asset(self, app_name: str, platform='Windows', update=False) -> AppAsset:
        """
        Returns an AppAsset object for the given app name and platform
        :param app_name: app name to get
        :param platform: platform to get asset for
        :param update: force update of asset list
        :return: AppAsset object
        :raises ValueError: if no asset is found for the given app name and platform
        """
        if update or platform not in self.uevmlfs.assets:
            self.get_assets(update_assets=True, platform=platform)

        try:
            return next(i for i in self.uevmlfs.assets[platform] if i.app_name == app_name)
        except StopIteration:
            raise ValueError

    def asset_available(self, item: App, platform='Windows') -> bool:
        """
        Returns whether an asset is available for the given item and platform
        :param item: item to check
        :param platform:
        :return: True if asset is available, False otherwise
        """
        # Just say yes for Origin titles
        if item.third_party_store:
            return True

        try:
            asset = self.get_asset(item.app_name, platform=platform)
            return asset is not None
        except ValueError:
            return False

    def get_item(self, app_name, update_meta=False, platform='Windows') -> App:
        """
        Returns an App object
        :param app_name: name to get
        :param update_meta: force update of metadata
        :param platform: platform to get app for
        :return: App object
        """
        if update_meta:
            self.get_asset_list(True, platform=platform)
        return self.uevmlfs.get_item_meta(app_name)

    def get_asset_list(self, update_assets=True, platform='Windows', filter_category='', force_refresh=False) -> (List[App], Dict[str, List[App]]):
        """
        Returns a list of all available assets for the given platform
        :param update_assets: force update of asset list
        :param platform: platform to get assets for
        :param filter_category: filter by category
        :param force_refresh: force refresh of asset list
        :return: Assets list
        """

        # Cancel all outstanding tasks and shut down the executor
        def stop_executor(tasks) -> None:
            """
            Cancel all outstanding tasks and shut down the executor
            :param tasks: tasks to cancel
            """
            for _, task in tasks.items():
                task.cancel()
            self.thread_executor.shutdown(wait=False)

        def fetch_asset_meta(name: str) -> bool:
            """
            Fetches asset metadata for the given app name and adds it to the list of assets
            :param name: app name
            :return: True if successful, False otherwise
            """
            if (name in currently_fetching or not fetch_list.get(name)) and ('Asset_Fetcher' in thread_enumerate()) or self.thread_executor_must_stop:
                return False

            thread_data = ''
            if self.use_threads:
                thread = current_thread()
                thread_data = f' ==> By Thread name={thread.name}'

            self.log.debug(f'--- START fetching data {name}{thread_data}')

            currently_fetching[name] = True
            start_time = datetime.now()
            name, namespace, catalog_item_id, _process_meta, _process_extras = fetch_list[name]

            if _process_meta:
                eg_meta, status_code = self.egs.get_item_info(namespace, catalog_item_id, timeout=10.0)
                if status_code != 200:
                    self.log.warning(f'Failed to fetch metadata for {name}: reponse code = {status_code}')
                    return False

                app = App(app_name=name, app_title=eg_meta['title'], metadata=eg_meta, asset_infos=assets[name])
                self.uevmlfs.set_item_meta(app.app_name, app)
                apps[name] = app

            if _process_extras:
                # we use title because it's less ambiguous than a name when searching an asset
                eg_extras = self.egs.get_assets_extras(asset_name=name, asset_title=apps[name].app_title, verbose_mode=self.verbose_mode)

                # check for data consistency
                if 'stomt' in app_name.lower() or 'terrainmagic' in app_name.lower():
                    if eg_extras.get('grab_result', '') != GrabResult.NO_ERROR.name or not eg_extras.get('purchased', False):
                        box_message(
                            msg=f'Some results in extras data are inconsistants for {app_name}. Please check the data and try again. Exiting...',
                            level='error'
                        )

                self.uevmlfs.set_item_extras(app_name=name, extras=eg_extras, update_global_dict=True)

                # log the asset if the title in metadata and the title in the marketplace grabbed page are not identical
                if eg_extras['page_title'] != '' and eg_extras['page_title'] != apps[name].app_title:
                    self.log.warning(f'{name} has incoherent data. It has been added to the bad_data_logger file')
                    eg_extras['grab_result'] = GrabResult.INCONSISTANT_DATA.name
                    if self.bad_data_logger:
                        self.bad_data_logger.info(name)

            # compute process time and average in s
            end_time = datetime.now()
            process_time = (end_time - start_time).total_seconds()
            self.process_time_average['time'] += process_time
            self.process_time_average['count'] += 1

            if fetch_list.get(name):
                del fetch_list[name]
                if self.verbose_mode:
                    self.log.info(f'Removed {name} from the metadata update')

            if currently_fetching.get(name):
                del currently_fetching[name]

            if not self.use_threads:
                process_average = self.process_time_average['time'] / self.process_time_average['count']
                self.log.info(f'===Time Average={process_average:.3f} s # ({(len(fetch_list) * process_average):.3f} s time left)')

            self.log.info(
                f'--- END fetching data in {name}{thread_data}. Time For Processing={process_time:.3f}s # Still {len(fetch_list)} assets to process'
            )
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                self.thread_executor_must_stop = True
                return False
            return True

        _ret = []
        meta_updated = False

        # fetch asset information for Windows, all installed platforms, and the specified one
        platforms = {'Windows'}
        platforms |= {platform}
        if gui_g.progress_window_ref is not None:
            gui_g.progress_window_ref.reset(new_value=0, new_text="Fetching platforms...", new_max_value=len(platforms))
        for _platform in platforms:
            self.get_assets(update_assets=update_assets, platform=_platform)
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                return []

        if not self.uevmlfs.assets:
            return _ret

        assets = {}
        if gui_g.progress_window_ref is not None:
            gui_g.progress_window_ref.reset(new_value=0, new_text="Fetching assets...", new_max_value=len(self.uevmlfs.assets.items()))
        for _platform, _assets in self.uevmlfs.assets.items():
            for _asset in _assets:
                if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                    return []
                if _asset.app_name in assets:
                    assets[_asset.app_name][_platform] = _asset
                else:
                    assets[_asset.app_name] = {_platform: _asset}

        fetch_list = {}
        assets_bypassed = {}
        apps = {}

        # loop through assets items to check for if they are for ue or not
        valid_items = []
        bypass_count = 0
        self.log.info(f'======\nSTARTING phase 1: asset indexing (ue or not)\n')
        if gui_g.progress_window_ref is not None:
            gui_g.progress_window_ref.reset(new_value=0, new_text="Indexing assets...", new_max_value=len(assets.items()))
        # note: we sort by reverse, as it the most recent version of an asset will be listed first
        for app_name, app_assets in sorted(assets.items(), reverse=True):
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                return []
            # notes:
            #   asset_id is not unique because somme assets can have the same asset_id but with several UE versions
            #   app_name is unique because it includes the unreal version
            # asset_id = app_assets['Windows'].asset_id
            assets_bypassed[app_name] = False
            if app_assets['Windows'].namespace != 'ue':
                self.log.debug(f'{app_name} has been bypassed (namespace != "ue") in phase 1')
                bypass_count += 1
                assets_bypassed[app_name] = True
                continue

            item = {'name': app_name, 'asset': app_assets}
            valid_items.append(item)

        self.ue_assets_count = len(valid_items)

        self.log.info(f'A total of {bypass_count} on {len(valid_items)} assets have been bypassed in phase 1')

        # check if we must refresh ue asset metadata cache
        self.check_for_ue_assets_updates(self.ue_assets_count, force_refresh)
        force_refresh = self.cache_is_invalidate
        if force_refresh:
            self.log.info(f'!! Assets metadata will be updated !!\n')
        else:
            self.log.info(f"Asset metadata won't be updated\n")

        self.log.info(f'======\nSTARTING phase 2:asset filtering and metadata updating\n')
        if gui_g.progress_window_ref is not None:
            gui_g.progress_window_ref.reset(new_value=0, new_text="Updating metadata...", new_max_value=len(valid_items))
        # loop through valid items to check for update and filtering
        bypass_count = 0
        filtered_items = []
        currently_fetching = {}
        i = 0
        while i < len(valid_items):
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                return []
            item = valid_items[i]
            app_name = item['name']
            app_assets = item['asset']
            if self.verbose_mode:
                self.log.info(f'Checking {app_name}....')

            item_metadata = self.uevmlfs.get_item_meta(app_name)
            asset_updated = False

            if not item_metadata:
                self.log.info(f'Metadata for {app_name} are missing. It Will be ADDED to the FETCH list')
            else:
                category = str(item_metadata.metadata['categories'][0]['path']).lower()
                if filter_category and filter_category.lower() not in category:
                    self.log.info(
                        f'{app_name} has been FILTERED by category ("{filter_category}" text not found in "{category}").It has been added to the ignored_logger file'
                    )
                    if self.ignored_logger:
                        self.ignored_logger.info(app_name)
                    assets_bypassed[app_name] = True
                    bypass_count += 1
                    i += 1
                    continue
                asset_updated = any(item_metadata.app_version(_p) != app_assets[_p].build_version for _p in app_assets.keys())
                apps[app_name] = item_metadata
                self.log.debug(f'{app_name} has been ADDED to the apps list with asset_updated={asset_updated}')

            # get extras data only in not filtered
            if force_refresh or asset_updated:
                process_extras = True
            else:
                # will read the extras data from file if necessary and put in the global dict
                process_extras = self.uevmlfs.get_item_extras(app_name) is None

            process_meta = not item_metadata or force_refresh or asset_updated

            if update_assets and (process_extras or process_meta):
                self.log.debug(f'Scheduling metadata and extras update for {app_name}')
                # namespace/catalog item are the same for all platforms, so we can just use the first one
                _ga = next(iter(app_assets.values()))
                fetch_list[app_name] = (app_name, _ga.namespace, _ga.catalog_item_id, process_meta, process_extras)
                meta_updated = True
            i += 1
            filtered_items.append(item)
            # end while i < len(valid_items):

        # setup and teardown of thread pool takes some time, so only do it when it makes sense.
        self.use_threads = len(fetch_list) > 5
        # self.use_threads = False  # test only
        if fetch_list:
            if gui_g.progress_window_ref is not None:
                gui_g.progress_window_ref.reset(
                    new_value=0, new_text="Fetching missing metadata...\nIt could take some time. Be patient.", new_max_value=len(fetch_list)
                )
                # gui_g.progress_window_ref.hide_progress_bar()
                # gui_g.progress_window_ref.hide_stop_button()

            self.log.info(f'Fetching metadata for {len(fetch_list)} app(s).')
            if self.use_threads:
                # note:  unreal engine API limits the number of connection to 16. So no more than 15 threads to avoid connection refused

                # with ThreadPoolExecutor(max_workers=min(16, os.cpu_count() - 2), thread_name_prefix="Asset_Fetcher") as executor:
                #    executor.map(fetch_asset_meta, fetch_list.keys(), timeout=30.0)
                self.thread_executor = ThreadPoolExecutor(max_workers=min(15, os.cpu_count() + 2), thread_name_prefix="Asset_Fetcher")
                # Dictionary that maps each key to its corresponding Future object
                futures = {}
                for key in fetch_list.keys():
                    # Submit the task and add its Future to the dictionary
                    future = self.thread_executor.submit(fetch_asset_meta, key)
                    futures[key] = future
                    if self.thread_executor_must_stop:
                        self.log.info(f'User stop has been pressed. Stopping running threads....')
                        stop_executor(futures)
                        return []

        self.log.info(f'A total of {bypass_count} on {len(valid_items)} assets have been bypassed in phase 2')
        self.log.info(f'======\nSTARTING phase 3: emptying the List of assets to be fetched \n')
        if gui_g.progress_window_ref is not None:
            # gui_g.progress_window_ref.show_progress_bar()  # show progress bar, must be before reset
            gui_g.progress_window_ref.show_stop_button()
            gui_g.progress_window_ref.reset(new_value=0, new_text="Checking and Fetching assets data...", new_max_value=len(filtered_items))
        # loop through valid and filtered items
        meta_updated = (bypass_count == 0) and meta_updated  # to avoid deleting metadata files or assets that have been filtered
        fetch_try_count = {}
        fetch_try_limit = 3
        while len(filtered_items) > 0:
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                return []
            item = filtered_items.pop()
            app_name = item['name']
            app_assets = item['asset']
            if fetch_try_count.get(app_name):
                fetch_try_count[app_name] += 1
            else:
                fetch_try_count[app_name] = 1
            if self.verbose_mode:
                self.log.info(f'Checking {app_name} Try number = {fetch_try_count[app_name] }. Still {len(filtered_items)} assets to check')
            try:
                app_item = apps.get(app_name)
            except (KeyError, IndexError):
                self.log.debug(f'{app_name} has not been found int the app list. Bypassing')
                # item not found in app, ignore and pass to next one
                continue
            # retry if the asset is still in fetch list (with active fetcher treads)
            if fetch_list.get(app_name) and (not currently_fetching.get(app_name) or 'Asset_Fetcher' not in thread_enumerate()):
                self.log.info(f'Fetching metadata for {app_name} is still no done, retrying')
                if currently_fetching.get(app_name):
                    del currently_fetching[app_name]
                fetch_asset_meta(app_name)

            if fetch_try_count[app_name] > fetch_try_limit:
                self.log.error(f'Fetching metadata for {app_name} has failed {fetch_try_limit} times. Skipping')
                continue
            try:
                is_bypassed = (app_name in assets_bypassed) and (assets_bypassed[app_name])
                is_a_mod = any(i['path'] == 'mods' for i in app_item.metadata.get('categories', []))
            except (KeyError, IndexError, AttributeError):
                self.log.debug(f'{app_name} has no metadata. Adding to the fetch list (again)')
                try:
                    fetch_list[app_name] = (app_name, item.namespace, item.catalog_item_id, True, True)
                    _ret.append(app_item)
                except (KeyError, IndexError, AttributeError):
                    self.log.debug(f'{app_name} has an invalid format. Could not been added to the fetch list')
                continue

            has_valid_platform = platform in app_assets
            is_still_fetching = (app_name in fetch_list) or (app_name in currently_fetching)

            if is_still_fetching:
                # put again the asset in the list waiting when it will be fetched
                filtered_items.append(item)
                time.sleep(3)  # Sleep for 3 seconds to let the fetch process progress or end

            # check if the asset will be added to the final list
            if not is_bypassed and not is_still_fetching and not is_a_mod and has_valid_platform:
                _ret.append(app_item)

        self.log.info(f'A total of {len(_ret)} assets have been analysed and kept in phase 3')

        self.update_aliases(force=meta_updated)

        if gui_g.s.never_update_data_files:
            meta_updated = False
        if meta_updated:
            if gui_g.progress_window_ref is not None:
                gui_g.progress_window_ref.reset(new_value=0, new_text="Updating metadata files...", new_max_value=len(_ret))
            # delete old files
            self._prune_metadata()
            self._save_metadata(_ret)
        if meta_updated:
            if gui_g.progress_window_ref is not None:
                gui_g.progress_window_ref.reset(new_value=0, new_text="Updating extras data files...", new_max_value=len(_ret))
            # save new ones
            self._prune_extras_data(update_global_dict=False)
            self._save_extras_data(self.uevmlfs.assets_extras_data, update_global_dict=False)
        return _ret

    # end def get_asset_list(self, update_assets=True, platform='Windows', filter_category='') -> (List[App], Dict[str, List[App]]):

    def _prune_metadata(self) -> None:
        """
        Compile a list of assets without assets, then delete their metadata
        """
        # compile list of assets without assets, then delete their metadata
        available_assets = set()
        available_assets |= {i.app_name for i in self.get_assets(platform='Windows')}

        for app_name in self.uevmlfs.get_item_app_names():
            self.log.debug(f'Removing old/unused metadata for "{app_name}"')
            self.uevmlfs.delete_item_meta(app_name)

    def _prune_extras_data(self, update_global_dict: True) -> None:
        """
        Compile a list of assets without assets, then delete their extras data
        :param update_global_dict:  if True, update the global dict
        """
        available_assets = set()
        available_assets |= {i.app_name for i in self.get_assets(platform='Windows')}

        for app_name in self.uevmlfs.get_item_app_names():
            self.log.debug(f'Removing old/unused extras data for "{app_name}"')
            self.uevmlfs.delete_item_extras(app_name, update_global_dict=update_global_dict)

    def _save_metadata(self, assets) -> None:
        """
        Save the metadata for the given assets
        :param assets:  List of assets to save
        """
        for app in assets:
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                return
            self.uevmlfs.set_item_meta(app.app_name, app)

    def _save_extras_data(self, extras: dict, update_global_dict: True) -> None:
        """
        Save the extras data for the given assets
        :param extras: Dict of extras data to save
        :param update_global_dict: if True, update the global dict
        """
        for app_name, eg_extras in extras.items():
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                return
            self.uevmlfs.set_item_extras(app_name=app_name, extras=eg_extras, update_global_dict=update_global_dict)

    def get_non_asset_library_items(self, force_refresh=False, skip_ue=True) -> (List[App], Dict[str, List[App]]):
        """
        Gets a list of Items without assets for installation, for instance Items delivered via
        third-party stores that do not have assets for installation

        :param force_refresh: Force a metadata refresh
        :param skip_ue: Ignore Unreal Marketplace entries
        :return: List of Items that do not have assets
        """
        _ret = []
        # get all the app names we have to ignore
        ignore = set(i.app_name for i in self.get_assets())

        for lib_item in self.egs.get_library_items():
            if lib_item['namespace'] == 'ue' and skip_ue:
                continue
            if lib_item['appName'] in ignore:
                continue

            item = self.uevmlfs.get_item_meta(lib_item['appName'])
            if not item or force_refresh:
                eg_meta, status_code = self.egs.get_item_info(lib_item['namespace'], lib_item['catalogItemId'])
                if status_code != 200:
                    self.log.warning(f'Failed to fetch metadata for {lib_item["appName"]}: reponse code = {status_code}')
                    continue
                item = App(app_name=lib_item['appName'], app_title=eg_meta['title'], metadata=eg_meta)
                self.uevmlfs.set_item_meta(item.app_name, item)

            if not any(i['path'] == 'mods' for i in item.metadata.get('categories', [])):
                _ret.append(item)

        # Force refresh to make sure these titles are included in aliasing
        self.update_aliases(force=True)
        return _ret

    @staticmethod
    def load_manifest(data: bytes) -> Manifest:
        """
        Load a manifest
        :param data: Bytes object to load the manifest from
        :return: Manifest object
        """
        if data[0:1] == b'{':
            return JSONManifest.read_all(data)
        else:
            return Manifest.read_all(data)

    def get_cdn_urls(self, item, platform='Windows'):
        """
        Get the CDN URLs
        :param item: Item to get the CDN URLs for
        :param platform: Platform to get the CDN URLs for
        :return: List of CDN URLs
        """
        m_api_r = self.egs.get_item_manifest(item.namespace, item.catalog_item_id, item.app_name, platform)

        # never seen this outside the launcher itself, but if it happens: PANIC!
        if len(m_api_r['elements']) > 1:
            raise ValueError('Manifest response has more than one element!')

        manifest_hash = m_api_r['elements'][0]['hash']
        base_urls = []
        manifest_urls = []
        for manifest in m_api_r['elements'][0]['manifests']:
            base_url = manifest['uri'].rpartition('/')[0]
            if base_url not in base_urls:
                base_urls.append(base_url)

            if 'queryParams' in manifest:
                params = '&'.join(f'{p["name"]}={p["value"]}' for p in manifest['queryParams'])
                manifest_urls.append(f'{manifest["uri"]}?{params}')
            else:
                manifest_urls.append(manifest['uri'])

        return manifest_urls, base_urls, manifest_hash

    def get_cdn_manifest(self, item, platform='Windows', disable_https=False):
        """
        Get the CDN manifest
        :param item: Item to get the CDN manifest for
        :param platform: Platform to get the CDN manifest for
        :param disable_https: Disable HTTPS for the manifest URLs
        :return: list of base URLs, manifest hash
        :raises ValueError
        """
        manifest_urls, base_urls, manifest_hash = self.get_cdn_urls(item, platform)
        if not manifest_urls:
            raise ValueError('No manifest URLs returned by API')

        if disable_https:
            manifest_urls = [url.replace('https://', 'http://') for url in manifest_urls]

        r = {}
        for url in manifest_urls:
            self.log.debug(f'Trying to download manifest from "{url}"...')
            try:
                r = self.egs.unauth_session.get(url, timeout=10.0)
            except Exception as error:
                self.log.warning(f'Unable to download manifest from "{urlparse(url).netloc}" '
                                 f'(Exception: {error!r}), trying next URL...')
                continue

            if r.status_code == 200:
                manifest_bytes = r.content
                break
            else:
                self.log.warning(f'Unable to download manifest from "{urlparse(url).netloc}" '
                                 f'(status: {r.status_code}), trying next URL...')
        else:
            raise ValueError(f'Unable to get manifest from any CDN URL, last result: {r.status_code} ({r.reason})')

        if sha1(manifest_bytes).hexdigest() != manifest_hash:
            raise ValueError('Manifest sha hash mismatch!')

        return manifest_bytes, base_urls

    def get_uri_manifest(self, uri: str) -> (bytes, List[str]):
        """
        Get the manifest
        :param uri: URI to get the manifest from
        :return:  Manifest data and base URLs
        """
        if uri.startswith('http'):
            r = self.egs.unauth_session.get(uri)
            r.raise_for_status()
            new_manifest_data = r.content
            base_urls = [r.url.rpartition('/')[0]]
        else:
            base_urls = []
            with open(uri, 'rb') as f:
                new_manifest_data = f.read()

        return new_manifest_data, base_urls

    # Check if the UE assets metadata cache must be updated
    def check_for_ue_assets_updates(self, assets_count: int, force_refresh=False) -> None:
        """
        Check if the UE assets metadata cache must be updated
        :param assets_count: assets count from the API
        :param force_refresh: force the refresh of the cache
        """
        self.cache_is_invalidate = False
        cached = self.uevmlfs.get_ue_assets_cache_data()
        cached_assets_count = cached['ue_assets_count']

        date_now = datetime.now().timestamp()
        date_diff = date_now - cached['last_update']

        if not cached_assets_count or cached_assets_count != assets_count:
            self.log.info(f'New assets are available. {assets_count} available VS {cached_assets_count} in cache')
            self.uevmlfs.set_ue_assets_cache_data(last_update=cached['last_update'], ue_assets_count=assets_count)

        if force_refresh or date_diff > self.ue_assets_max_cache_duration:
            self.cache_is_invalidate = True
            self.uevmlfs.set_ue_assets_cache_data(last_update=date_now, ue_assets_count=assets_count)
            if not force_refresh:
                self.log.info(f'Data cache is outdated. Cache age is {date_diff:.1f} s OR {str(timedelta(seconds=date_diff))}')
        else:
            self.log.info(f'Data cache is still valid. Cache age is {str(timedelta(seconds=date_diff))}')

    def clean_exit(self, code=0) -> None:
        """
        Do cleanup, config saving, and quit
        :param code: exit code
        """
        self.uevmlfs.save_config()
        logging.shutdown()
        sys.exit(code)



# coding=utf-8
"""
Implementation for:
- CSV_headings: contains the title of each column and a boolean value to know if its contents must be preserved if it already exists in the output file (To Avoid overwriting data changed by the user in the file)
"""
import uuid

import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience

CSV_headings = {
    'Asset_id': False,  # ! important: Do not Rename => this field is used as main key for each asset
    'App name': False,
    'App title': False,
    'Category': False,
    'Review': False,
    'Developer': False,
    'Description': False,
    'Status': False,
    'Discount price': False,
    'On sale': False,
    'Purchased': False,
    'Obsolete': False,
    'Supported versions': False,
    'Grab result': False,
    'Price': False,  # ! important: Rename Wisely => this field is searched by text in the next lines
    'Old price': False,  # ! important: always place it after the Price field in the list
    # User Fields
    'Comment': True,
    'Stars': True,
    'Must buy': True,
    'Test result': True,
    'Installed folder': True,
    'Alternative': True,
    'Origin': True,
    # less important fields
    'Page title': False,
    'Image': False,
    'Url': True,  # could be kept if a better url that can be used to download the asset is found
    'Compatible versions': False,
    'Date added': True,
    'Creation date': False,
    'Update date': False,
    'UE version': False,
    'Uid': False
}


def create_emty_csv_row(return_as_string=False):
    """
    Create an empty row for the csv file
    :return: empty row
    """
    data = {}
    for key, _ in CSV_headings.items():
        data[key] = 0  # 0 is used to avoid empty cells in the csv file
    data['Asset_id'] = 'dummy_row_' + str(uuid.uuid4())[:8]  # dummy unique Asset_id to avoid issue
    data['Image'] = gui_g.s.empty_cell  # avoid displaying image warning on mouse over

    if return_as_string:
        data = ','.join(str(value) for value in data.values())
    return data



# coding: utf-8
"""
Implementation for:
- EPCAPI : Epic Games Client API
- GrabResult : Enum for the result of grabbing a page.
- create_empty_assets_extras : Creates an empty asset extras dict.
"""
import logging
import re
import urllib.parse
from enum import Enum

import requests
import requests.adapters
from bs4 import BeautifulSoup
from requests.auth import HTTPBasicAuth

from UEVaultManager.models.exceptions import InvalidCredentialsError


class GrabResult(Enum):
    """
    Enum for the result of grabbing a page.
    """
    NO_ERROR = 0
    INCONSISTANT_DATA = 1
    PAGE_NOT_FOUND = 2
    CONTENT_NOT_FOUND = 3
    TIMEOUT = 4


def create_empty_assets_extras(asset_name: str) -> dict:
    """
    Creates an empty asset extras dict.
    :param asset_name:  The name of the asset.
    :return: The empty asset extras dict.
     """
    return {
        'asset_name': asset_name,
        'asset_url': '',
        'asset_name_in_url': '',
        'price': 0,
        'discount_price': 0,
        'review': 0,
        'purchased': False,
        'supported_versions': '',
        'page_title': '',
        'grab_result': GrabResult.NO_ERROR.name,
        'on_sale': False
    }


class EPCAPI:
    """
    Epic Games Client API
    :param lc: The language code.
    :param cc: The country code.
    :param timeout: The timeout for requests.
    """
    _user_agent = 'UELauncher/11.0.1-14907503+++Portal+Release-Live Windows/10.0.19041.1.256.64bit'
    _store_user_agent = 'EpicGamesLauncher/14.0.8-22004686+++Portal+Release-Live'
    # required for the oauth request
    _user_basic = '34a02cf8f4414e29b15921876da36f9a'
    _pw_basic = 'daafbccc737745039dffe53d94fc76cf'
    _label = 'Live-EternalKnight'

    _oauth_host = 'account-public-service-prod03.ol.epicgames.com'
    _launcher_host = 'launcher-public-service-prod06.ol.epicgames.com'
    _entitlements_host = 'entitlement-public-service-prod08.ol.epicgames.com'
    _catalog_host = 'catalog-public-service-prod06.ol.epicgames.com'
    _ecommerce_host = 'ecommerceintegration-public-service-ecomprod02.ol.epicgames.com'
    _datastorage_host = 'datastorage-public-service-liveegs.live.use1a.on.epicgames.com'
    _library_host = 'library-service.live.use1a.on.epicgames.com'
    # Using the actual store host with a user-agent newer than 14.0.8 leads to a CF verification page,
    # but the dedicated graphql host works fine.
    # _store_gql_host = 'launcher.store.epicgames.com'
    _store_gql_host = 'graphql.epicgames.com'
    _artifact_service_host = 'artifact-public-service-prod.beee.live.use1a.on.epicgames.com'
    _search_url = 'www.unrealengine.com/marketplace/en-US'
    _login_url = 'www.unrealengine.com/id/login/epic'
    ignored_logger = None

    def __init__(self, lc='en', cc='US', timeout=10.0):
        self.log = logging.getLogger('EPCAPI')
        self.notfound_logger = None  # will be setup when created in core.py
        self.session = requests.session()
        self.session.headers['User-Agent'] = self._user_agent
        # increase maximum pool size for multithreaded metadata requests
        self.session.mount('https://', requests.adapters.HTTPAdapter(pool_maxsize=16))

        self.unauth_session = requests.session()
        self.unauth_session.headers['User-Agent'] = self._user_agent

        self._oauth_basic = HTTPBasicAuth(self._user_basic, self._pw_basic)

        self.access_token = None
        self.user = None

        self.language_code = lc
        self.country_code = cc

        self.request_timeout = timeout if timeout > 0 else None

    def _extract_price_from_elt(self, dom_elt=None, asset_name='NO NAME') -> float:
        """
        Extracts the price from a BeautifulSoup element.
        :param dom_elt: The BeautifulSoup element.
        :param asset_name: The name of the asset (for display purpose only).
        :return: The price.
        """
        if dom_elt is None or dom_elt == '':
            self.log.debug(f'Price not found for {asset_name}')
            return 0.0
        price = 0.0
        try:
            # asset base price when logged
            price = dom_elt.text.strip('$')
            price = price.strip('€')
            price = float(price)
        except Exception as error:
            self.log.warning(f'Can not find the price for {asset_name}:{error!r}')
        return price

    def get_auth_url(self) -> str:
        """
        Returns the url for the oauth login.
        :return: The url
        """
        login_url = 'https://www.epicgames.com/id/login?redirectUrl='
        redirect_url = f'https://www.epicgames.com/id/api/redirect?clientId={self._user_basic}&responseType=code'
        return login_url + urllib.parse.quote(redirect_url)

    def resume_session(self, session: dict) -> dict:
        """
        Resumes a session.
        :param session: The session.
        :return: The session.
        """
        self.session.headers['Authorization'] = f'bearer {session["access_token"]}'
        r = self.session.get(f'https://{self._oauth_host}/account/api/oauth/verify', timeout=self.request_timeout)
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorMessage' in j:
            self.log.warning(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])

        # update other data
        session.update(j)
        self.user = session
        return self.user

    def start_session(
        self, refresh_token: str = None, exchange_token: str = None, authorization_code: str = None, client_credentials: bool = False
    ) -> dict:
        """
        Starts a session.
        :param refresh_token: refresh token
        :param exchange_token: exchange token
        :param authorization_code: authorization code
        :param client_credentials: client credentials
        :return: The session.
        :raise: ValueError,InvalidCredentialsError
        """
        if refresh_token:
            params = dict(grant_type='refresh_token', refresh_token=refresh_token, token_type='eg1')
        elif exchange_token:
            params = dict(grant_type='exchange_code', exchange_code=exchange_token, token_type='eg1')
        elif authorization_code:
            params = dict(grant_type='authorization_code', code=authorization_code, token_type='eg1')
        elif client_credentials:
            params = dict(grant_type='client_credentials', token_type='eg1')
        else:
            raise ValueError('At least one token type must be specified!')

        r = self.session.post(
            f'https://{self._oauth_host}/account/api/oauth/token', data=params, auth=self._oauth_basic, timeout=self.request_timeout
        )
        # Only raise HTTP exceptions on server errors
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorCode' in j:
            if j['errorCode'] == 'errors.com.epicgames.oauth.corrective_action_required':
                self.log.error(f'{j["errorMessage"]} ({j["correctiveAction"]}), '
                               f'open the following URL to take action: {j["continuationUrl"]}')
            else:
                self.log.error(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])
        elif r.status_code >= 400:
            self.log.error(f'EGS API responded with status {r.status_code} but no error in response: {j}')
            raise InvalidCredentialsError('Unknown error')

        self.session.headers['Authorization'] = f'bearer {j["access_token"]}'
        # only set user info when using non-anonymous login
        if not client_credentials:
            self.user = j

        return j

    def get_item_token(self) -> str:
        """
        Gets the item token.
        Unused but kept for the global API reference.
        :return: The item token using json format
        """
        r = self.session.get(f'https://{self._oauth_host}/account/api/oauth/exchange', timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_ownership_token(self, namespace: str, catalog_item_id: str) -> bytes:
        """
        Gets the ownership token.
        Unused but kept for the global API reference.
        :param namespace:  namespace
        :param catalog_item_id: catalog item id
        :return: The ownership token.
        """
        user_id = self.user.get('account_id')
        r = self.session.post(
            f'https://{self._ecommerce_host}/ecommerceintegration/api/public/'
            f'platforms/EPIC/identities/{user_id}/ownershipToken',
            data=dict(nsCatalogItemId=f'{namespace}:{catalog_item_id}'),
            timeout=self.request_timeout
        )
        r.raise_for_status()
        return r.content

    def get_external_auths(self) -> dict:
        """
        Gets the external auths.
        Unused but kept for the global API reference.
        :return: The external auths using json format.
        """
        user_id = self.user.get('account_id')
        r = self.session.get(f'https://{self._oauth_host}/account/api/public/account/{user_id}/externalAuths', timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_item_assets(self, platform='Windows', label='Live'):
        """
        Gets the item assets.
        :param platform: platform to get assets for
        :param label: label of the assets
        :return: The item assets using json format.
        """
        r = self.session.get(
            f'https://{self._launcher_host}/launcher/api/public/assets/{platform}', params=dict(label=label), timeout=self.request_timeout
        )
        r.raise_for_status()
        return r.json()

    def get_item_manifest(self, namespace, catalog_item_id, app_name, platform='Windows', label='Live') -> dict:
        """
        Gets the item manifest.
        :param namespace:  namespace
        :param catalog_item_id: catalog item id
        :param app_name: app name
        :param platform: platform to get manifest for
        :param label: label of the manifest
        :return: The item manifest using json format.
        """
        r = self.session.get(
            f'https://{self._launcher_host}/launcher/api/public/assets/v2/platform'
            f'/{platform}/namespace/{namespace}/catalogItem/{catalog_item_id}/app'
            f'/{app_name}/label/{label}',
            timeout=self.request_timeout
        )
        r.raise_for_status()
        return r.json()

    def get_launcher_manifests(self, platform='Windows', label: str = None) -> dict:
        """
        Gets the launcher manifests.
        Unused but kept for the global API reference.
        :param platform: platform to get manifests for
        :param label: label of the manifests
        :return: The launcher manifests using json format.
        """
        r = self.session.get(
            f'https://{self._launcher_host}/launcher/api/public/assets/v2/platform/'
            f'{platform}/launcher',
            timeout=self.request_timeout,
            params=dict(label=label if label else self._label)
        )
        r.raise_for_status()
        return r.json()

    def get_user_entitlements(self) -> dict:
        """
        Gets the user entitlements.
        Unused but kept for the global API reference.
        :return: The user entitlements using json format.
        """
        user_id = self.user.get('account_id')
        r = self.session.get(
            f'https://{self._entitlements_host}/entitlement/api/account/{user_id}/entitlements',
            params=dict(start=0, count=5000),
            timeout=self.request_timeout
        )
        r.raise_for_status()
        return r.json()

    def get_item_info(self, namespace: str, catalog_item_id: str, timeout: float = None) -> (dict, int):
        """
        Gets the item info.
        :param namespace: Namespace of the item
        :param catalog_item_id: Catalog item id of the item
        :param timeout: Timeout for the request
        :return: (The item info, status code)
        """
        r = self.session.get(
            f'https://{self._catalog_host}/catalog/api/shared/namespace/{namespace}/bulk/items',
            params=dict(
                id=catalog_item_id, includeDLCDetails=True, includeMainGameDetails=True, country=self.country_code, locale=self.language_code
            ),
            timeout=timeout or self.request_timeout
        )
        r.raise_for_status()
        return r.json().get(catalog_item_id, None), r.status_code

    def get_artifact_service_ticket(self, sandbox_id: str, artifact_id: str, label='Live', platform='Windows') -> dict:
        """
        Gets the artifact service ticket.
        Unused but kept for the global API reference.
        :param sandbox_id: sandbox id
        :param artifact_id: artifact id
        :param label: label
        :param platform: platform to get ticket for
        :return: The artifact service ticket using json format.
        """
        # Based on EOS Helper Windows service implementation. Only works with anonymous EOS Helper session.
        # sandbox_id is the same as the namespace, artifact_id is the same as the app name
        r = self.session.post(
            f'https://{self._artifact_service_host}/artifact-service/api/public/v1/dependency/'
            f'sandbox/{sandbox_id}/artifact/{artifact_id}/ticket',
            json=dict(label=label, expiresInSeconds=300, platform=platform),
            params=dict(useSandboxAwareLabel='false'),
            timeout=self.request_timeout
        )
        r.raise_for_status()
        return r.json()

    def get_item_manifest_by_ticket(self, artifact_id: str, signed_ticket: str, label='Live', platform='Windows') -> dict:
        """
        Gets the item manifest by ticket.
        Unused but kept for the global API reference.
        :param artifact_id: artifact id
        :param signed_ticket: signed ticket
        :param label: the label
        :param platform: platform to get manifest for
        :return: The item manifest by ticket using json format.
        """
        # Based on EOS Helper Windows service implementation.
        r = self.session.post(
            f'https://{self._launcher_host}/launcher/api/public/assets/v2/'
            f'by-ticket/app/{artifact_id}',
            json=dict(platform=platform, label=label, signedTicket=signed_ticket),
            timeout=self.request_timeout
        )
        r.raise_for_status()
        return r.json()

    def get_library_items(self, include_metadata=True) -> list:
        """
        Gets the library items.
        :param include_metadata: Whether to include metadata
        :return: The library items
        """
        records = []
        r = self.session.get(
            f'https://{self._library_host}/library/api/public/items', params=dict(includeMetadata=include_metadata), timeout=self.request_timeout
        )
        r.raise_for_status()
        j = r.json()
        records.extend(j['records'])

        # Fetch remaining library entries as long as there is a cursor
        while cursor := j['responseMetadata'].get('nextCursor', None):
            r = self.session.get(
                f'https://{self._library_host}/library/api/public/items',
                params=dict(includeMetadata=include_metadata, cursor=cursor),
                timeout=self.request_timeout
            )
            r.raise_for_status()
            j = r.json()
            records.extend(j['records'])

        return records

    def find_asset_url(self, asset_name: str, timeout=10.0) -> []:
        """
        Find the asset url from the asset name by searching the asset name in the unreal engine marketplace
        :param asset_name: asset name to search
        :param timeout: timeout for the request
        :return: (The asset url, the asset name , the grab result code)
        """
        # remove the suffix _EngineVersion (ex _4.27) at the end of the name to have a valid search value
        regex = r"_[4|5]\.\d{1,2}$"
        converted_name = re.sub(regex, '', asset_name, 0)
        # Replace ' ' by '%20'
        converted_name = converted_name.replace(' ', '%20')
        # SnakeCase
        # converted_name = inflection.underscore(converted_name)
        # Lower case
        converted_name = converted_name.lower()
        # Replace '_' by '-'
        # converted_name = converted_name.replace('_', '-')

        # remove some not alphanumeric cars (NOT ALL, keep %)
        entry_list = [':', ',', '.', ';', '=', '?', '!', '#', "/", "$", "€"]
        for entry in entry_list:
            converted_name = converted_name.replace(entry, '')

        url = ''
        asset_name_in_url = ''
        search_url_root = f'https://{self._search_url}/assets?keywords='
        search_url_full = search_url_root + converted_name
        try:
            r = self.session.get(search_url_full, timeout=timeout)
        except requests.exceptions.Timeout:
            self.log.warning(f'Timeout for {asset_name}')
            return [url, asset_name_in_url, GrabResult.TIMEOUT.name]
        if not r.ok:
            self.log.warning(f'Can not find the url for {asset_name}:{r.reason}')
            return [url, asset_name_in_url, GrabResult.PAGE_NOT_FOUND.name]

        soup = BeautifulSoup(r.content, 'html.parser')
        links = []
        group_elt = soup.find('div', attrs={'class': 'asset-list-group'})

        if "No content found" in group_elt.getText():
            self.log.info(f'{asset_name} has not been not found in marketplace.It has been added to the notfound_logger file')
            if self.ignored_logger:
                self.ignored_logger.info(asset_name)
            return [url, asset_name_in_url, GrabResult.CONTENT_NOT_FOUND.name]

        # find all links to assets that correspond to the search
        for link in group_elt.findAll('a', attrs={'class': 'mock-ellipsis-item mock-ellipsis-item-helper ellipsis-text'}):
            links.append(link.get('href'))

        # return the first one (probably the best choice)
        asset_name_in_url = links[0].replace('/marketplace/en-US/product/', '')
        url = 'https://www.unrealengine.com' + links[0]
        return [url, asset_name_in_url, GrabResult.NO_ERROR.name]

    #  get the extras data of an asset (price, review...)
    def get_assets_extras(self, asset_name: str, asset_title: str, timeout=10.0, verbose_mode=False) -> dict:
        """
        Get the extras data of an asset (price, review...)
        :param asset_name: name of the asset
        :param asset_title: title of the asset
        :param timeout: connection timeout
        :param verbose_mode: verbose mode
        :return: a dict with the extras data
        """
        not_found_price = 0.0
        not_found_review = 0.0
        supported_versions = ''
        page_title = ''
        no_result = create_empty_assets_extras(asset_name=asset_name)

        # try to find the url of the asset by doing a search in the marketplace
        asset_url, asset_name_in_url, error_code = self.find_asset_url(asset_title, timeout)
        if asset_url == '' or asset_name_in_url == '':
            no_result['grab_result'] = error_code
            return no_result
        try:
            response = self.session.get(asset_url)  # when using session, we are already logged in Epic game
            response.raise_for_status()
            self.log.info(f'Grabbing extras data for {asset_name}')
        except requests.exceptions.RequestException as error:
            self.log.warning(f'Can not get extras data for {asset_name}:{error!r}')
            no_result['grab_result'] = error_code
            return no_result

        soup_logged = BeautifulSoup(response.text, 'html.parser')

        price = not_found_price
        discount_price = not_found_price

        search_for_price = True

        # check if already purchased
        purchased = False
        purchased_elt = soup_logged.find('div', class_='purchase')
        if purchased_elt is not None:
            if 'Free' in purchased_elt.getText():
                # free price when logged
                price = 0.0
                search_for_price = False
                if verbose_mode:
                    self.log.info(f'{asset_name} is free (check 1)')
            elif 'Open in Launcher' in purchased_elt.getText():
                # purchased asset
                purchased = True
                if verbose_mode:
                    self.log.info(f'{asset_name} as already been purchased')
                # grab the price on a non logged soup (price will be available on that page only)
                try:
                    response = requests.get(asset_url)  # not using session, so not logged in Epic game
                    response.raise_for_status()
                    soup_not_logged = BeautifulSoup(response.text, 'html.parser')
                    purchased_elt = soup_not_logged.find('div', class_='purchase')
                except requests.exceptions.RequestException:
                    pass

        if search_for_price and purchased_elt is not None:
            if 'Sign in to Download' in purchased_elt.getText():
                # free price when logged or not
                price = 0.0
                if verbose_mode:
                    self.log.info(f'{asset_name} is free (check 2)')
            else:
                # get price using the logged or the not logged soup
                # notes:
                #   when not discounted
                #       base-price is not available
                #       price is 'save-discount'
                #       discount-price is 0.0
                #   when discounted
                #       price is 'base-price'
                #       discount-price is 'save-discount'
                elt = purchased_elt.find('span', class_='save-discount')
                current_price = self._extract_price_from_elt(elt, asset_name)
                elt = purchased_elt.find('span', class_='base-price')
                base_price = self._extract_price_from_elt(elt, asset_name)
                if elt is not None:
                    # discounted
                    price = base_price
                    discount_price = current_price
                else:
                    # not discounted
                    price = current_price
                    discount_price = current_price

        # get review
        reviews_elt = soup_logged.find('div', class_='asset-detail-rating')
        if reviews_elt is not None:
            reviews_elt = reviews_elt.find('div', class_='rating-board__pop__title')
        if reviews_elt is not None:
            try:
                inner_span_elt = reviews_elt.find('span')
                content = inner_span_elt.text
                pos = content.index(' out of ')
                review = float(content[0:pos])
            except Exception as error:
                self.log.warning(f'Can not find the review for {asset_name}:{error!r}')
                review = not_found_review
        else:
            self.log.debug(f'reviews not found for {asset_name}')
            review = not_found_review

        # get Supported Engine Versions
        title_elt = soup_logged.find('span', class_='ue-version')
        if title_elt is not None:
            try:
                supported_versions = title_elt.text
            except Exception as error:
                self.log.warning(f'Can not find the Supported Engine Versions for {asset_name}:{error!r}')
        else:
            self.log.debug(f'Can not find the Supported Engine Versions for {asset_name}')
            review = not_found_review

        # get page title
        title_elt = soup_logged.find('h1', class_='post-title')
        if title_elt is not None:
            try:
                page_title = title_elt.text
            except Exception as error:
                self.log.warning(f'Can not find the Page title for {asset_name}:{error!r}')
        else:
            self.log.debug(f'Can not find the Page title not found for {asset_name}')
            review = not_found_review
        on_sale = 0.0 < discount_price < price and price > 0.0

        self.log.info(f'GRAB results: asset_name_in_url={asset_name_in_url} on_sale={on_sale} purchased={purchased} price={price} review={review}')
        return {
            'asset_name': asset_name,
            'asset_url': asset_url,
            'asset_name_in_url': asset_name_in_url,
            'page_title': page_title,
            'price': price,
            'discount_price': discount_price,
            'review': review,
            'purchased': purchased,
            'supported_versions': supported_versions,
            'grab_result': error_code,
            'on_sale': on_sale
        }



# coding: utf-8
"""
Imlementations for:
- UpdateSeverity: Enum for update severity.
- UEVMAPI: Class for interacting with the UEVaultManager API.
- extract_codename and extract_severity functions
"""
import logging
import re
from enum import Enum
from platform import system

import requests
from packaging import version

# noinspection PyPep8Naming
from UEVaultManager import __version__ as UEVM_version


class UpdateSeverity(Enum):
    """
    Enum for update severity.
    """
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


def extract_codename(pypi_description: str) -> str:
    """
    Extracts the codename from the PyPI description.
    :param pypi_description: PyPI's description
    :return: codename
    """
    pattern = r'## codename:(.*?)##'
    match = re.search(pattern, pypi_description, re.IGNORECASE | re.DOTALL)
    if match:
        return match.group(1).strip()
    else:
        return ''


def extract_severity(old_version: str, new_version: str):
    """
    Gets the update severity of a new version
    :param old_version: old version in standard format "major.minor.micro"
    :param new_version: new version in standard format "major.minor.micro"
    :return: update severity (UpdateSeverity type)
    """
    old_ver = version.parse(old_version)
    new_ver = version.parse(new_version)

    if new_ver > old_ver:
        if new_ver.major > old_ver.major:
            return UpdateSeverity.HIGH
        elif new_ver.minor > old_ver.minor:
            return UpdateSeverity.MEDIUM
        elif new_ver.micro > old_ver.micro:
            return UpdateSeverity.LOW
    else:
        return UpdateSeverity.NONE


class UEVMAPI:
    """
    Class for interacting with the UEVaultManager API.
    """
    _package_name = 'UEVaultManager'
    _user_agent = f'{_package_name}/{UEVM_version} ({system()})'

    def __init__(self):
        self.session = requests.session()
        self.log = logging.getLogger('UEVMAPI')
        self.session.headers['User-Agent'] = self._user_agent

    def get_version_information(self) -> dict:
        """
        Gets the latest packagage version information from PyPI.
        :return: version information
        """
        url = f'https://pypi.org/pypi/{self._package_name}/json'
        r = self.session.get(url, timeout=10.0)
        r.raise_for_status()
        data = r.json()
        pypi_version = '0.0.0'
        pypi_codename = ''
        severity = UpdateSeverity.NONE
        release_url = ''
        summary = ''
        try:
            pypi_version = data['info']['version']
            pypi_codename = extract_codename(data['info']['description'])
            severity = str(extract_severity(UEVM_version, pypi_version).name)
            release_url = data['info']['release_url']
            summary = data['info']['summary']
        except KeyError:
            self.log.warning('Failed to get version information from PyPI.')

        return dict(version=pypi_version, codename=pypi_codename, severity=severity, release_url=release_url, summary=summary)



# coding=utf-8
"""
Implementation for:
- SaferDict: a dictionary subclass that provides a safer alternative to handle non-existing keys
"""


class SaferDict(dict):
    """
    A dictionary subclass that provides a safer alternative to handle non-existing keys.
    It returns None or the specified default value when accessing non-existing keys, rather than raising a KeyError.
    """

    def __getitem__(self, key):
        """
        Returns the value associated with the given key.
        If the key does not exist, returns None.
        :param key: the key to get
        """
        return super().get(key, None)

    def __getattr__(self, key):
        """
        Returns the value associated with the given key using dot notation.
        If the key does not exist, returns None.
        :param key: the key to get
        """
        return super().get(key, None)

    def __delattr__(self, key):
        """
        Deletes the key-value pair associated with the given key using dot notation.
        If the key does not exist, does nothing and does not raise an error.
        :param key: the key to delete
        """
        self.pop(key, None)

    def __setattr__(self, key, value):
        """
        Sets the value associated with the given key using dot notation.
        :param key: the key to set
        """
        self[key] = value

    def get(self, key: str, default=None):
        """
        Returns the value associated with the given key.
        If the key does not exist, returns the specified default value or None if no default value is provided.
        :param key: the key to get
        :param default: the default value to return if the key doesn't exist
        """
        return super().get(key, default)

    def copy_from(self, source: dict):
        """
        Copies the content of the given source dictionary into the SaferDict.
        :param source: the source dictionary to copy from
        :raises TypeError: if the source is not a dictionary
        """
        if not isinstance(source, dict):
            raise TypeError("source must be a dictionary")
        self.clear()
        self.update({k: v for k, v in source.items()})



# coding=utf-8
"""
Implementation for:
- ProgressWindow: window to display the progress of a function
"""
import queue
import threading
import tkinter as tk
from tkinter import ttk

import UEVaultManager.tkgui.modules.functions as gui_f  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.functions_no_deps as gui_fn  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience


# noinspection PyProtectedMember
class ProgressWindow(tk.Toplevel):
    """
    The window to display the progress of a function
    :param title: the title
    :param width: the width
    :param height: the height
    :param icon: the icon
    :param screen_index: the index of the screen on which the window will be displayed
    :param max_value: the maximum value of the progress bar
    :param show_start_button: whether to show the start button
    :param show_stop_button: whether to show the stop button
    :param show_progress: whether to show the progress bar
    :param function: the function to execute
    :param function_parameters: the parameters of the function
    :param quit_on_close: whether to quit the application when the window is closed
    """

    def __init__(
        self,
        title: str,
        width: 300,
        height: 150,
        icon=None,
        screen_index=0,
        max_value=100,
        show_start_button=True,
        show_stop_button=True,
        show_progress=True,
        function=None,
        function_parameters: dict = None,
        quit_on_close=False
    ):
        super().__init__()
        self.title(title)
        geometry = gui_fn.center_window_on_screen(screen_index, height, width)
        self.geometry(geometry)
        gui_fn.set_icon_and_minmax(self, icon)
        self._thread_check_delay = 100
        self.must_end = False
        self.quit_on_close = quit_on_close
        self.max_value = max_value
        self.continue_execution = True
        self.function = function
        self.function_params = function_parameters
        self.function_return_value = None
        self.result_queue = queue.Queue()

        self.content_frame = self.ContentFrame(self)
        self.content_frame.pack(ipadx=5, ipady=5, padx=5, pady=5, fill=tk.X)

        if not show_start_button and not show_stop_button:
            self.control_frame = None
        else:
            self.control_frame = self.ControlFrame(self)
            self.control_frame.pack(ipadx=5, ipady=5, padx=5, pady=5, fill=tk.X)

        gui_g.progress_window_ref = self

        if not show_progress:
            self.hide_progress_bar()
        if self.control_frame and not show_start_button:
            self.control_frame.button_start.pack_forget()
        if self.control_frame and not show_stop_button:
            self.control_frame.button_stop.pack_forget()

        # Start the execution if not control frame is present
        # important because the control frame is not present when the function is set after the window is created
        if self.control_frame is None or not show_start_button:
            self.start_execution()

    def __del__(self):
        gui_f.log_debug(f'Destruction of {self.__class__.__name__} object')
        gui_g.progress_window_ref = None

    class ContentFrame(ttk.Frame):
        """
        The frame that contains the content of the window
        :param container: the container
        """

        def __init__(self, container):
            super().__init__(container)
            pack_def_options = {'ipadx': 3, 'ipady': 3, 'padx': 5, 'pady': 5, 'fill': tk.X}
            if container.function is None:
                lbl_function = ttk.Label(self, text='No callable function set Yet')
            else:
                lbl_function = ttk.Label(self, text='Running function: ' + container.function.__name__)
            lbl_function.pack(**pack_def_options)
            progress_bar = ttk.Progressbar(self, orient="horizontal", mode="determinate", maximum=container.max_value)
            progress_bar.pack(**pack_def_options)
            self.pack_def_options = pack_def_options
            self.progress_bar = progress_bar
            self.lbl_function = lbl_function

    class ControlFrame(ttk.Frame):
        """
        The frame that contains the control buttons
        :param container: the container
        :param show_start_button: whether to show the start button
        :param show_stop_button: whether to show the stop button
        """

        def __init__(self, container, show_start_button=True, show_stop_button=True):
            super().__init__(container)
            pack_def_options = {'ipadx': 3, 'ipady': 3, 'fill': tk.X}
            self.pack_def_options = pack_def_options
            self.button_start = None
            self.button_stop = None
            if show_start_button:
                button_start = ttk.Button(self, text="Start", command=container.start_execution)
                button_start.pack(**pack_def_options, side=tk.LEFT)
                self.button_start = button_start
            if show_stop_button:
                button_stop = ttk.Button(self, text="Stop", command=container.stop_execution, state=tk.DISABLED)
                button_stop.pack(**pack_def_options, side=tk.RIGHT)
                self.button_stop = button_stop

    def _function_result_wrapper(self, function, *args, **kwargs) -> None:
        """
        Wraps the function call and puts the result in the queue
        :param function: the function to execute
        :param args: args to pass to the function
        :param kwargs: kwargs to pass to the function
        """
        gui_f.log_info(f'execution of {function.__name__} has started')
        result = function(*args, **kwargs)
        gui_f.log_info(f'execution of {function.__name__} has ended')
        self.result_queue.put(result)

    def _check_for_end(self, t: threading) -> None:
        """
        Checks if the thread has ended, if not, schedules another check
        :param t: the thread to check
        """
        if t.is_alive():
            # Schedule another check in a few ms
            self.after(self._thread_check_delay, self._check_for_end, t)
        else:
            # The thread has finished; clean up
            gui_f.log_debug(f'Quitting {self.__class__.__name__}')
            self.function_return_value = self.result_queue.get()
            self.close_window(destroy_window=self.quit_on_close)  # the window is kept to allow further calls to the progress bar

    def set_text(self, new_text: str) -> None:
        """
        Sets the text of the label
        :param new_text: the new text
        """
        if self.content_frame is None or self.content_frame.lbl_function is None:
            return
        self.content_frame.lbl_function.config(text=new_text)

    def set_value(self, new_value: int) -> None:
        """
        Sets the value of the progress bar
        :param new_value: the new value
        """
        new_value = max(0, new_value)
        if self.content_frame is None or self.content_frame.progress_bar is None:
            return
        self.content_frame.progress_bar["value"] = new_value

    def set_max_value(self, new_max_value: int) -> None:
        """
        Sets the maximum value of the progress bar
        :param new_max_value: the new maximum value
        """
        if self.content_frame is None or self.content_frame.progress_bar is None:
            return
        self.max_value = new_max_value
        self.content_frame.progress_bar["maximum"] = new_max_value

    def set_function(self, new_function) -> None:
        """
        Sets the function to execute
        :param new_function: the new function
        """
        if new_function is None:
            return
        self.set_text('Running function: ' + new_function.__name__)
        self.function = new_function

    def set_function_parameters(self, parameters: dict) -> None:
        """
        Sets the parameters to pass to the function
        :param parameters: the parameters
        """
        self.function_params = parameters

    def hide_progress_bar(self) -> None:
        """
        Hides the progress bar
        """
        self.content_frame.progress_bar.pack_forget()

    def show_progress_bar(self) -> None:
        """
        Shows the progress bar
        """
        self.content_frame.progress_bar.pack(self.content_frame.pack_def_options)

    def hide_start_button(self) -> None:
        """
        Hides the start button
        """
        if self.control_frame is None or self.control_frame.button_start is None:
            return
        self.control_frame.button_start.pack_forget()

    def show_start_button(self) -> None:
        """
        Shows the start button
        """
        if self.control_frame is None or self.control_frame.button_start is None:
            return
        self.control_frame.button_start.pack(self.control_frame.pack_def_options)

    def hide_stop_button(self) -> None:
        """
        Hides the stop button
        """
        if self.control_frame is None or self.control_frame.button_stop is None:
            return
        self.control_frame.button_stop.pack_forget()

    def show_stop_button(self) -> None:
        """
        Shows the stop button
        """
        if self.control_frame is None or self.control_frame.button_stop is None:
            return
        self.control_frame.button_stop.pack(self.control_frame.pack_def_options)

    def reset(self, new_title=None, new_value=None, new_text=None, new_max_value=None) -> None:
        """
        Resets the progress bar
        :param new_title: the new title
        :param new_value: the new value
        :param new_text: the new text
        :param new_max_value: the new maximum value
        """
        self.must_end = False
        try:
            # sometimes the window is already destroyed
            if new_title is not None:
                self.title(new_title)
            if new_text is not None:
                self.set_text(new_text)
            if new_value is not None:
                self.show_progress_bar()
                self.set_value(new_value)
            if new_max_value is not None:
                self.show_progress_bar()
                self.set_max_value(new_max_value)
        except tk.TclError:
            gui_f.log_warning('Some tkinter elements are not set. The window is probably already destroyed')
        self.continue_execution = True

    def start_execution(self) -> None:
        """
        Starts the execution of the function
        """
        if self.function is None:
            gui_f.log_warning('the function name to execute is not set')
            return
        self.continue_execution = True
        self.deactivate()
        # Run the function in a separate thread
        t = threading.Thread(target=self._function_result_wrapper, args=(self.function, self), kwargs=self.function_params)
        t.start()
        # Schedule GUI update while waiting for the thread to finish
        self.after(self._thread_check_delay, self._check_for_end, t)

    def stop_execution(self) -> None:
        """
        Stops the execution of the function
        """
        self.continue_execution = False
        self.activate()

    def get_result(self):
        """
        Returns the result of the function
        """
        return self.function_return_value

    def activate(self) -> None:
        """
        Activates the control buttons for starting
        """
        if self.control_frame is None:
            return
        if self.control_frame.button_start is not None:
            self.control_frame.button_start.config(state=tk.NORMAL)
        if self.control_frame.button_stop is not None:
            self.control_frame.button_stop.config(state=tk.DISABLED)

    def deactivate(self) -> None:
        """
        Activates the control buttons for stopping
        """
        if self.control_frame is None:
            return
        if self.control_frame.button_start is not None:
            self.control_frame.button_stop.config(state=tk.NORMAL)
        if self.control_frame.button_stop is not None:
            self.control_frame.button_start.config(state=tk.DISABLED)

    def update_and_continue(self, value=0, increment=0) -> bool:
        """
        Updates the progress bar and returns whether the execution should continue
        :param value: the value to set
        :param increment: the value to increment. If both value and increment are set, the value is ignored
        """
        try:
            # sometimes the window is already destroyed
            progress_bar = self.content_frame.progress_bar
            if increment:
                value = progress_bar["value"] + increment

            if value > self.max_value:
                value = self.max_value
            progress_bar["value"] = value
            progress_bar.update_idletasks()
        except tk.TclError:
            gui_f.log_warning('Some tkinter elements are not set. The window is probably already destroyed')
        self.update_idletasks()
        return self.continue_execution

    def close_window(self, destroy_window=True, _event=None) -> None:
        """
        Closes the window
        :param destroy_window: whether to destroy the window or just hide it
        :param _event: the event that triggered the close
        """
        self.must_end = True
        if destroy_window:
            if self.quit_on_close:
                self.quit()
            else:
                self.destroy()



# coding=utf-8
"""
Implementation for:
- DisplayContentWindow: the window to display a text content
"""
import os
import tkinter as tk
from tkinter import filedialog as fd
from tkinter import ttk

from ttkbootstrap.constants import *

import UEVaultManager.tkgui.modules.functions as gui_f  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.functions_no_deps as gui_fn  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
from UEVaultManager.tkgui.modules.ExtendedWidgetClasses import ExtendedText


# class DisplayContentWindow(tk.Tk if tk._default_root is None else tk.Toplevel):
class DisplayContentWindow(tk.Toplevel):
    """
    Window to display a text content
    :param title: the title of the window
    :param width: the width of the window
    :param height: the height of the window
    :param icon: the icon of the window
    :param screen_index: the index of the screen on which the window will be displayed
    :param quit_on_close: whether to quit the application when the window is closed

    """

    def __init__(self, title: str, width: int = 600, height: int = 430, icon=None, screen_index=0, quit_on_close=False):
        super().__init__()
        self.title(title)
        style = gui_fn.set_custom_style(gui_g.s.theme_name, gui_g.s.theme_font)
        self.style = style
        # if tk._default_root == self :
        #     style = gui_f.set_custom_style(gui_g.s.theme_name, gui_g.s.theme_font)
        #     print(f"style SELF = {style.theme_use()}")
        # else:
        #     style = ttk.Style(tk._default_root)
        #     print(f"style ROOT = {style.theme_use()}")
        #
        # self.style = style

        geometry = gui_fn.center_window_on_screen(screen_index, height, width)
        self.geometry(geometry)
        gui_fn.set_icon_and_minmax(self, icon)
        self.resizable(True, False)
        self.quit_on_close = quit_on_close
        self.keep_existing = False  # whether to keep the existing content when adding a new one
        self.content_frame = self.ContentFrame(self)
        self.control_frame = self.ControlFrame(self)

        self.content_frame.pack(ipadx=5, ipady=5, padx=5, pady=5, fill=tk.X)
        self.control_frame.pack(ipadx=5, ipady=5, padx=5, pady=5, fill=tk.X)

        self.bind('<Key>', self.on_key_press)
        self.protocol("WM_DELETE_WINDOW", self.on_close)

        gui_g.display_content_window_ref = self

    class ContentFrame(ttk.Frame):
        """
        The frame containing the content of the window
        :param container: the container of the frame
        """

        def __init__(self, container):
            super().__init__(container)
            pack_def_options = {'ipadx': 3, 'ipady': 3}

            text_content = ExtendedText(self)
            scrollbar = ttk.Scrollbar(self)
            scrollbar.config(command=text_content.yview)
            text_content.config(yscrollcommand=scrollbar.set)
            scrollbar.pack(side=tk.RIGHT, fill=tk.Y, **pack_def_options)
            text_content.pack(side=tk.LEFT, fill=tk.BOTH, expand=True, **pack_def_options)

            self.text_content = text_content

    class ControlFrame(ttk.Frame):
        """
        The frame containing the control buttons of the window
        :param container: the container of the frame
        """

        def __init__(self, container):
            super().__init__(container)
            pack_def_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.BOTH, 'expand': False}
            lblf_def_options = {'ipadx': 1, 'ipady': 1, 'expand': True, 'fill': tk.X}
            lblf_commands = ttk.LabelFrame(self, text='Commands')
            lblf_commands.pack(**lblf_def_options)
            # noinspection PyArgumentList
            ttk.Button(lblf_commands, text='Clean content', command=container.clean).pack(**pack_def_options, side=tk.LEFT)
            # noinspection PyArgumentList
            ttk.Button(lblf_commands, text='Save To File', command=container.save_to_file).pack(**pack_def_options, side=tk.LEFT)
            # noinspection PyArgumentList
            ttk.Button(lblf_commands, text='Close', command=container.on_close, bootstyle=WARNING).pack(**pack_def_options, side=tk.RIGHT)

    def on_key_press(self, event) -> None:
        """
        Event when a key is pressed
        :param event: the event that triggered the call of this function
        """
        if event.keysym == 'Escape':
            self.on_close()

    def on_close(self, _event=None) -> None:
        """
        Event when the window is closing
        :param _event: the event that triggered the call of this function
        """
        self.close_window()

    def close_window(self) -> None:
        """
        Close the window
        """
        gui_g.display_content_window_ref = None
        if self.quit_on_close:
            self.quit()
        else:
            self.destroy()

    def display(self, content='', keep_mode=True) -> None:
        """
        Display the content in the window. By default, ie. keep_mode==True, each new call adds the content to the existing content with a new line.
        :param content: the text to print
        :param keep_mode: whether to keep the existing content when a new one is added
        """

        if self.keep_existing:
            content += '\n'
            self.content_frame.text_content.insert(tk.END, content)
        else:
            self.content_frame.text_content.delete('1.0', tk.END)
            self.content_frame.text_content.insert(tk.END, content)
        # set the mode at the end to allow using display() to be used to change the mode for the next call
        self.keep_existing = keep_mode

    def clean(self) -> None:
        """
        Clean the content of the window
        """
        self.content_frame.text_content.delete('1.0', tk.END)

    def save_to_file(self) -> str:
        """
        Save the content displayed to a file
        """
        initial_dir = os.path.dirname(gui_g.s.csv_filename)
        filename = fd.asksaveasfilename(
            title='Choose a file to save data to', initialdir=initial_dir, filetypes=gui_g.s.data_filetypes, initialfile='UEVM_output.txt'
        )
        if filename:
            with open(filename, 'w') as f:
                f.write(self.content_frame.text_content.get('1.0', tk.END))
            gui_f.box_message(f'Content Saved to {filename}')
        return filename



# coding=utf-8
"""
Implementation for:
- EditableTable is a custom class that extends the pandastable.Table class, providing additional functionality
"""
import io
import webbrowser
from tkinter import ttk

import pandas as pd
from pandas.errors import EmptyDataError
from pandastable import Table, TableModel, config

from UEVaultManager.models.csv import create_emty_csv_row, CSV_headings
from UEVaultManager.tkgui.modules.EditCellWindowClass import EditCellWindow
from UEVaultManager.tkgui.modules.EditRowWindowClass import EditRowWindow
from UEVaultManager.tkgui.modules.ExtendedWidgetClasses import ExtendedText, ExtendedCheckButton
from UEVaultManager.tkgui.modules.functions import *
from UEVaultManager.tkgui.modules.functions_no_deps import convert_to_bool
from UEVaultManager.tkgui.modules.TaggedLabelFrameClass import TaggedLabelFrame


class EditableTable(Table):
    """
    EditableTable is a custom class that extends the pandastable.Table class, providing additional functionality
    such as loading data from CSV files, searching, filtering, pagination, and editing cell values.
    :param container_frame: The parent frame for the table.
    :param file: The path to the CSV file containing the table data.
    :param fontsize: The font size for the table.
    :param show_toolbar: Whether to show the toolbar.
    :param show_statusbar: Whether to show the status bar.
    :param kwargs: Additional arguments to pass to the pandastable.Table class.
    """

    def __init__(self, container_frame=None, file=None, rows_per_page=36, show_toolbar=False, show_statusbar=False, **kwargs):
        self._last_selected_row = -1
        self._last_selected_col = -1

        self.file = file
        self.must_save = False
        self.must_rebuild = False
        self.container_frame = container_frame
        self.show_toolbar = show_toolbar
        self.show_statusbar = show_statusbar

        self.pagination_enabled = True
        self.rows_per_page = rows_per_page
        self.current_page = 0
        self.total_pages = 0

        self.data = None
        self.data_filtered = None

        self.edit_row_window = None
        self.edit_row_entries = None
        self.edit_row_index = None

        self.edit_cell_window = None
        self.edit_cell_row_index = None
        self.edit_cell_col_index = None
        self.edit_cell_entry = None

        self.load_data()
        Table.__init__(self, container_frame, dataframe=self.data, showtoolbar=show_toolbar, showstatusbar=show_statusbar, **kwargs)
        self.bind('<Double-Button-1>', self.create_edit_cell_window)

    def _generate_cell_selection_changed_event(self) -> None:
        """
        Creates the event bindings for the table.
        """
        selected_row = self.currentrow
        selected_col = self.currentcol

        if (selected_row != self._last_selected_row) or (selected_col != self._last_selected_col):
            self._last_selected_row = selected_row
            self._last_selected_col = selected_col
            self.event_generate('<<CellSelectionChanged>>')

    def gradient_color_cells(self, col_names=None, cmap='sunset', alpha=1) -> None:
        """
        Creates a gradient color for the cells os specified columns. The gradient depends on the cell value between min and max values for that column.
        :param col_names: The names of the columns to create a gradient color for.
        :param cmap: name of the colormap to use
        :param alpha: alpha value for the color
        """
        # import pylab as plt
        # cmaps = sorted(m for m in plt.cm.datad if not m.endswith("_r"))
        # print(cmaps)
        # possible cmaps:
        # 'Accent', 'Blues', 'BrBG', 'BuGn', 'BuPu', 'CMRmap', 'Dark2', 'GnBu', 'Greens', 'Greys', 'OrRd', 'Oranges', 'PRGn', 'Paired', 'Pastel1',
        #  'Pastel2', 'PiYG', 'PuBu', 'PuBuGn', 'PuOr', 'PuRd', 'Purples', 'RdBu', 'RdGy', 'RdPu', 'RdYlBu', 'RdYlGn', 'Reds', 'Set1', 'Set2', 'Set3',
        #  'Spectral', 'Wistia', 'YlGn', 'YlGnBu', 'YlOrBr', 'YlOrRd', 'afmhot', 'autumn', 'binary', 'bone', 'brg', 'bwr', 'cool', 'coolwarm', 'copper',
        #  'cubehelix', 'flag', 'gist_earth', 'gist_gray', 'gist_heat', 'gist_ncar', 'gist_rainbow', 'gist_stern', 'gist_yarg', 'gnuplot', 'gnuplot2',
        #  'gray', 'hot', 'hsv', 'jet', 'nipy_spectral', 'ocean', 'pink', 'prism', 'rainbow', 'seismic', 'spring', 'summer', 'tab10', 'tab20', 'tab20b',
        #  'tab20c', 'terrain', 'winter'

        if col_names is None:
            return
        df = self.data_filtered if self.data_filtered is not None else self.model.df
        for colname in col_names:
            x = df[colname]
            clrs = self.values_to_colors(x, cmap, alpha)
            clrs = pd.Series(clrs, index=df.index)
            rc = self.rowcolors
            rc[colname] = clrs

    def color_cells_if(self, col_names=None, color='green', value_to_check='True') -> None:
        """
        Set the cell color for the specified columns and the cell with a given value.
        :param col_names: The names of the columns to create a gradient color for.
        :param color: The color to set the cell to.
        :param value_to_check: The value to check for.
        """

        if col_names is None:
            return
        df = self.data_filtered if self.data_filtered is not None else self.model.df
        for col_name in col_names:
            if col_name not in df.columns:
                continue
            mask = df[col_name] == value_to_check
            self.setColorByMask(col=col_name, mask=mask, clr=color)

    def color_cells_if_not(self, col_names=None, color='grey', value_to_check='False') -> None:
        """
        Set the cell color for the specified columns and the cell with NOT a given value.
        :param col_names: The names of the columns to create a gradient color for.
        :param color: The color to set the cell to.
        :param value_to_check: The value to check for.
        """
        if col_names is None:
            return
        df = self.data_filtered if self.data_filtered is not None else self.model.df
        for col_name in col_names:
            if col_name not in df.columns:
                continue
            mask = df[col_name] != value_to_check
            self.setColorByMask(col=col_name, mask=mask, clr=color)

    def color_rows_if(self, col_names=None, color='#55555', value_to_check='True') -> None:
        """
        Set the row color for the specified columns and the rows with a given value.
        :param col_names: The names of the columns to check for the value.
        :param color: The color to set the row to.
        :param value_to_check: The value to check for.
        """
        if col_names is None:
            return
        df = self.data_filtered if self.data_filtered is not None else self.model.df

        for col_name in col_names:
            if col_name not in df.columns:
                continue
            row_indices = []
            mask = df[col_name]
            for i in range(min(self.rows_per_page, len(mask))):
                try:
                    if str(mask[i]) == value_to_check:
                        row_indices.append(i)
                except KeyError:
                    log_debug(f'KeyError for row {i} in color_rows_if')
            if len(row_indices) > 0:  # Check if there are any row indices
                try:
                    self.setRowColors(rows=row_indices, clr=color, cols='all')
                except KeyError:
                    log_debug(f'KeyError for row in color_rows_if')
            return

    def set_preferences(self, default_pref=None) -> None:
        """
        Initializes the table preferences.
        :param default_pref: The default preferences to apply to the table.
        """
        # remove the warning: "A value is trying to be set on a copy of a slice from a DataFrame"
        # when sorting the table with pagination enabled
        # see: https://stackoverflow.com/questions/20625582/how-to-deal-with-settingwithcopywarning-in-pandas
        pd.options.mode.chained_assignment = None

        if default_pref is not None:
            config.apply_options(default_pref, self)

    def colorRows(self):
        """
        Color individual cells in column(s). Requires that the rowcolors
        dataframe has been set. This needs to be updated if the index is reset
        Override this method to check indexes when rebuildind data from en empty table
        """
        df = self.model.df
        rc = self.rowcolors
        rows = self.visiblerows
        offset = rows[0]
        idx = df.index[rows]

        for col in self.visiblecols:
            colname = df.columns[col]
            if colname in list(rc.columns):
                try:
                    colors = rc[colname].loc[idx]
                except KeyError:
                    colors = None
                if colors is not None:
                    for row in rows:
                        clr = colors.iloc[row - offset]
                        if not pd.isnull(clr):
                            self.drawRect(row, col, color=clr, tag='colorrect', delete=0)

    def set_colors(self) -> None:
        """
        Initializes the colors of some cells depending on their values.
        """
        if not gui_g.s.use_colors_for_data:
            self.redraw()
            return
        log_debug('set_colors')
        self.gradient_color_cells(col_names=['Review'], cmap='Set3', alpha=1)
        self.color_cells_if(col_names=['Purchased', 'On sale'], color='lightgreen', value_to_check='True')
        self.color_cells_if(col_names=['Grab result'], color='lightblue', value_to_check='NO_ERROR')
        self.color_cells_if_not(col_names=['Status'], color='#555555', value_to_check='ACTIVE')
        self.color_rows_if(col_names=['Status'], color='#555555', value_to_check='SUNSET')
        self.color_rows_if(col_names=['Obsolete'], color='#777777', value_to_check='True')
        self.redraw()

    def handle_left_click(self, event) -> None:
        """
        Handles left-click events on the table.
        :param event: The event that triggered the function call.
        """
        super().handle_left_click(event)
        self._generate_cell_selection_changed_event()

    def handle_right_click(self, event) -> None:
        """
        Handles right-click events on the table.
        :param event: The event that triggered the function call.
        """
        super().handle_right_click(event)
        self._generate_cell_selection_changed_event()

    def show_page(self, page=None) -> None:
        """
        Displays the specified page of the table data.
        :param page: The page number to display (zero-based index).
        """
        if page is None:
            page = self.current_page
        if self.pagination_enabled:
            self.total_pages = min(self.total_pages, len(self.data))
            if page < 0:
                page = 0
            elif page >= self.total_pages:
                page = self.total_pages - 1
            # Calculate start and end rows for current page
            self.current_page = page
            start = page * self.rows_per_page
            end = start + self.rows_per_page
            start = min(start, len(self.data))
            end = min(end, len(self.data))
            try:
                # Update table with data for current page
                self.model.df = self.data.iloc[start:end]
            except AttributeError:
                # self.redraw()
                log_debug('AttributeError in show_page')
                self.set_colors()
                return
        else:
            # Update table with all data
            self.model.df = self.data_filtered
            self.current_page = 0
        # self.redraw()
        self.set_colors()

    def next_page(self) -> None:
        """
        Navigates to the next page of the table data.
        """
        if self.current_page <= self.total_pages:
            self.show_page(self.current_page + 1)

    def prev_page(self) -> None:
        """
        Navigates to the previous page of the table data.
        """
        if self.current_page > 0:
            self.show_page(self.current_page - 1)

    def first_page(self) -> None:
        """
        Navigates to the first page of the table data.
        """
        self.show_page(0)

    def last_page(self) -> None:
        """
        Navigates to the last page of the table data.
        """
        self.show_page(self.total_pages - 1)

    def load_data(self) -> None:
        """
        Loads data from the specified CSV file into the table.
        """
        csv_options = {
            'converters': {
                'Asset_id': str,  #
                'App name': str,  #
                'Review': float,  #
                'Price': float,  #
                'Old price': float,  #
                'Purchased': convert_to_bool,  #
                'Must buy': convert_to_bool,  #
                'Date added': convert_to_datetime,  #
            },
            'on_bad_lines': 'warn',
            'encoding': "utf-8",
        }
        if self.file is None or not os.path.isfile(self.file):
            log_warning(f'File to read data from is not defined or not found: {self.file}')
            return

        self.must_rebuild = False
        try:
            self.data = pd.read_csv(self.file, **csv_options)
        except EmptyDataError:
            # will create an empty row with the correct columns
            str_data = ','.join(str(value) for value in CSV_headings.keys())  # column names
            str_data += '\n'
            str_data += create_emty_csv_row(return_as_string=True)  # dummy row
            self.data = pd.read_csv(io.StringIO(str_data))
            self.must_rebuild = True

        self.init_data_format()

        # log_debug("\nCOL TYPES AFTER LOADING CSV\n")
        # self.data.info()  # direct print info

        self.total_pages = (len(self.data) - 1) // self.rows_per_page + 1
        self.data_filtered = self.data

    def init_data_format(self) -> None:
        """
        Initializes the data format for the table.
        """
        for col in self.data.columns:
            try:
                self.data[col] = self.data[col].astype(str)
            except (KeyError, ValueError) as error:
                log_error(f'Could not convert column "{col}" to string. Error: {error}')

        col_to_convert = ['Creation date', 'Update date']
        # note "date added" does not use the same format as the other date columns
        for col in col_to_convert:
            col_type = self.data.get(col, None)
            if col_type is not None:
                try:
                    self.data[col] = pd.to_datetime(self.data[col], format='ISO8601', errors='ignore')
                except (KeyError, ValueError) as error:
                    log_error(f'Could not convert column "{col}" to datetime. Error: {error}')

        col_to_convert = ['Review', 'Price', 'Old price', 'Discount price']
        for col in col_to_convert:
            col_type = self.data.get(col, None)
            if col_type is not None:
                try:
                    self.data[col] = self.data[col].astype(float)
                except (KeyError, ValueError) as error:
                    log_error(f'Could not convert column "{col}" to float. Error: {error}')

        col_to_convert = ['Category', 'Grab result']
        for col in col_to_convert:
            col_type = self.data.get(col, None)
            if col_type is not None:
                try:
                    self.data[col] = self.data[col].astype('category')
                except (KeyError, ValueError) as error:
                    log_error(f'Could not convert column "{col}" to category. Error: {error}')

        # log_debug("\nCOL TYPES AFTER MANUAL CONVERSION\n")
        # self.data.info() # direct print info

        self.data_filtered = self.data

    def reload_data(self) -> None:
        """
        Reloads data from the CSV file and refreshes the table display.
        """
        self.load_data()
        self.show_page(self.current_page)

    def rebuild_data(self) -> bool:
        """
         Rebuilds the data in the table based on the current filtering and sorting options.
         :return: True if the data was successfully rebuilt, False otherwise.
         """

        # we use a string comparison here to avoid to import of the module to check the real class of UEVM_cli_ref
        if gui_g.UEVM_cli_ref is None or 'UEVaultManagerCLI' not in str(type(gui_g.UEVM_cli_ref)):
            from_cli_only_message()
            return False
        else:
            gui_g.UEVM_cli_ref.list_assets(gui_g.UEVM_cli_args)
            self.current_page = 0
            self.load_data()
            self.show_page(0)
            return True

    def save_data(self) -> None:
        """
        Saves the current table data to the CSV file.
        """

        data = self.data.iloc[0:len(self.data)]
        self.updateModel(TableModel(data))  # needed to restore all the data and not only the current page
        self.model.df.to_csv(self.file, index=False, na_rep='N/A', date_format=gui_g.s.csv_datetime_format)
        self.show_page(self.current_page)
        self.must_save = False

    def apply_filters(self, filter_dict=None, global_search=None) -> None:
        """
        Searches the table data based on the provided filter dictionary.
        :param filter_dict: A dictionary containing the column names as keys and filter values as the values.
        :param global_search: The text to search for in the table data.
        """
        if filter_dict is None:
            filter_dict = {}
        log_info(f'Filtering data with: {filter_dict} and global search: {global_search}')
        self.data_filtered = self.data

        for key, value in filter_dict.items():
            if key == 'Category' and value == gui_g.s.default_category_for_all:
                continue
            if key == 'Grab result' and value == gui_g.s.default_category_for_all:
                continue
            self.data_filtered = self.data_filtered[self.data_filtered[key].apply(lambda x: str(value).lower() in str(x).lower())]

        if global_search and global_search != gui_g.s.default_global_search:
            self.data_filtered = self.data_filtered[self.data_filtered.apply(lambda row: global_search.lower() in str(row).lower(), axis=1)]
        self.show_page(0)

    def reset_filters(self) -> None:
        """
        Resets the table data filtering and sorting, displaying all rows.
        """
        self.data_filtered = self.data
        self.show_page(0)

    def expand_columns(self) -> None:
        """
        Expands the width of all columns in the table.
        """
        self.expandColumns(factor=gui_g.s.expand_columns_factor)

    def contract_columns(self) -> None:
        """
        Contracts the width of all columns in the table.
        """
        self.contractColumns(factor=gui_g.s.contract_columns_factor)

    def autofit_columns(self) -> None:
        """
        Automatically resizes the columns to fit their content.
        """
        # note:
        # autoResizeColumns() will not resize table with more than 30 columns
        # same limit without settings the limit in adjustColumnWidths()
        # self.autoResizeColumns()
        self.adjustColumnWidths(limit=len(self.data.columns))
        self.redraw()

    def zoom_in(self) -> None:
        """
        Increases the font size of the table.
        """
        self.zoomIn()

    def zoom_out(self) -> None:
        """
        Decreases the font size of the table.
        """
        self.zoomOut()

    def get_edited_row_values(self) -> dict:
        """
        Returns the values of the selected row in the table.
        :return: A dictionary containing the column names and their corresponding values for the selected row.
        """
        if self.edit_row_entries is None or self.edit_row_index is None:
            return {}
        entries_values = {}
        for key, entry in self.edit_row_entries.items():
            try:
                value = entry.get()
            except AttributeError:
                value = entry.get_content()  # for extendedWidgets
            except TypeError:
                value = entry.get('1.0', 'end')
            entries_values[key] = value
        return entries_values

    def create_edit_record_window(self) -> None:
        """
        Creates the edit row window for the selected row in the table.
        """
        row_selected = self.getSelectedRow()
        if row_selected is None:
            return

        title = 'Edit current row'
        width = 900
        height = 1000
        # window is displayed at mouse position
        # x = self.master.winfo_rootx()
        # y = self.master.winfo_rooty()
        edit_row_window = EditRowWindow(
            parent=self.master, title=title, width=width, height=height, icon=gui_g.s.app_icon_filename, editable_table=self
        )
        edit_row_window.grab_set()
        edit_row_window.minsize(width, height)
        # configure the grid
        edit_row_window.content_frame.columnconfigure(0, weight=0)
        edit_row_window.content_frame.columnconfigure(1, weight=1)

        self.edit_record(row_selected)

    def edit_record(self, row_selected: int = None) -> None:
        """
        Edits the values of the specified row in the table.
        :param row_selected: The index of the row to edit.
        """
        edit_row_window = gui_g.edit_row_window_ref
        if row_selected is None or edit_row_window is None:
            return
        # get and display the row data
        row_data = self.model.df.iloc[row_selected].to_dict()
        entries = {}
        image_url = ''
        for i, (key, value) in enumerate(row_data.items()):
            from tkinter import ttk
            ttk.Label(edit_row_window.content_frame, text=key).grid(row=i, column=0, sticky=tk.W)
            lower_key = key.lower()
            if lower_key == 'image':
                image_url = value

            if lower_key == 'asset_id':
                # asset_id is readonly
                entry = ttk.Entry(edit_row_window.content_frame)
                entry.insert(0, value)
                entry.grid(row=i, column=1, sticky=tk.EW)
            elif lower_key == 'url':
                # we add a button to open the url in an inner frame
                inner_frame_url = tk.Frame(edit_row_window.content_frame)
                inner_frame_url.grid(row=i, column=1, sticky=tk.EW)
                entry = ttk.Entry(inner_frame_url)
                entry.insert(0, value)
                entry.pack(side=tk.LEFT, fill=tk.X, expand=True)
                button = ttk.Button(inner_frame_url, text="Open URL", command=self.open_asset_url)
                button.pack(side=tk.RIGHT)
            elif lower_key in ('description', 'comment'):
                entry = ExtendedText(edit_row_window.content_frame, height=3)
                entry.set_content(value)
                entry.grid(row=i, column=1, sticky=tk.EW)
            elif lower_key in ('must buy', 'obsolete', 'purchased', 'on sale'):
                entry = ExtendedCheckButton(edit_row_window.content_frame, label='', images_folder=gui_g.s.assets_folder)
                entry.set_content(value)
                entry.grid(row=i, column=1, sticky=tk.EW)
            else:
                # other field is just a usual entry
                entry = ttk.Entry(edit_row_window.content_frame)
                entry.insert(0, value)
                entry.grid(row=i, column=1, sticky=tk.EW)

            entries[key] = entry

        # image preview
        show_asset_image(image_url=image_url, canvas_image=edit_row_window.control_frame.canvas_image)

        self.edit_row_entries = entries
        self.edit_row_index = row_selected
        self.edit_row_window = edit_row_window
        edit_row_window.initial_values = self.get_edited_row_values()

    def save_edit_row_record(self) -> None:
        """
        Saves the edited row values to the table data.
        """
        for key, value in self.get_edited_row_values().items():
            self.model.df.at[self.edit_row_index, key] = value
        self.edit_row_entries = None
        self.edit_row_index = None
        self.redraw()
        self.must_save = True
        self.edit_row_window.close_window()
        self.update_quick_edit()

    def move_to_prev_record(self) -> None:
        """
        Navigates to the previous row in the table and opens the edit row window.
        """
        row_selected = self.getSelectedRow()
        if row_selected is None or row_selected == 0:
            return
        self.setSelectedRow(row_selected - 1)
        self.redraw()
        self._generate_cell_selection_changed_event()
        self.edit_record(row_selected - 1)

    def move_to_next_record(self) -> None:
        """
        Navigates to the next row in the table and opens the edit row window.
        """
        row_selected = self.getSelectedRow()
        if row_selected is None or row_selected == self.model.df.shape[0] - 1:
            return
        self.setSelectedRow(row_selected + 1)
        self.redraw()
        self._generate_cell_selection_changed_event()
        self.edit_record(row_selected + 1)

    def create_edit_cell_window(self, event) -> None:
        """
        Creates the edit cell window for the selected cell in the table.
        :param event: The event that triggered the creation of the edit cell window.
        """
        row_index = self.get_row_clicked(event)
        col_index = self.get_col_clicked(event)
        if row_index is None or col_index is None:
            return None
        cell_value = self.model.df.iat[row_index, col_index]

        title = 'Edit current cell values'
        width = 300
        height = 110
        # window is displayed at mouse position
        # x = self.master.winfo_rootx()
        # y = self.master.winfo_rooty()
        edit_cell_window = EditCellWindow(parent=self.master, title=title, width=width, height=height, editable_table=self)
        edit_cell_window.grab_set()
        edit_cell_window.minsize(width, height)

        # get and display the cell data
        col_name = self.model.df.columns[col_index]
        ttk.Label(edit_cell_window.content_frame, text=col_name).pack(side=tk.LEFT)
        entry = ttk.Entry(edit_cell_window.content_frame)
        entry.insert(0, cell_value)
        entry.pack(side=tk.LEFT, fill=tk.X, expand=True)
        entry.focus_set()

        self.edit_cell_entry = entry
        self.edit_cell_row_index = row_index
        self.edit_cell_col_index = col_index
        self.edit_cell_window = edit_cell_window
        edit_cell_window.initial_values = self.get_edit_cell_values()

    def get_edit_cell_values(self) -> str:
        """
        Returns the values of the selected cell in the table.
        :return: The value of the selected cell.
        """
        if self.edit_cell_entry is None:
            return ''
        return self.edit_cell_entry.get()

    def save_edit_cell_value(self) -> None:
        """
        Saves the edited cell value to the table data.
        """
        widget = self.edit_cell_entry
        if widget is None or self.edit_cell_row_index is None or self.edit_cell_col_index is None or self.edit_cell_entry is None:
            return
        # widget_class = widget.winfo_class().lower()
        entry = self.edit_cell_entry
        try:
            try:
                # get value for an entry tk widget
                value = entry.get()
            except TypeError:
                # get value for a text tk widget
                value = entry.get('1.0', 'end')
            self.model.df.iat[self.edit_cell_row_index, self.edit_cell_col_index] = value
            self.must_save = True
            self.edit_cell_entry = None
            self.edit_cell_row_index = None
            self.edit_cell_col_index = None
        except TypeError:
            log_warning(f'Failed to get content of {widget}')
        self.redraw()
        self.edit_cell_window.close_window()
        self.update_quick_edit()

    def update_quick_edit(self, quick_edit_frame: TaggedLabelFrame = None, row: int = None) -> None:
        """
        Quick edit the content some cells of the selected row.
        :param quick_edit_frame: The frame to display the cell content preview in.
        :param row: The row index of the selected cell.
        """
        if row is None or row >= len(self.model.df) or quick_edit_frame is None:
            return

        column_names = ['Url', 'Comment', 'Stars', 'Test result', 'Must buy', 'Alternative', 'Installed folder', 'Origin']
        for col_name in column_names:
            col = self.model.df.columns.get_loc(col_name)
            value = self.model.getValueAt(row=row, col=col)
            quick_edit_frame.set_child_values(tag=col_name, content=value, row=row, col=col)

    @staticmethod
    def quick_edit(quick_edit_frame: TaggedLabelFrame = None) -> None:
        """
        Resets the cell content preview.
        :param quick_edit_frame: The frame to reset the cell content preview in.
        """
        quick_edit_frame.set_default_content('Url')
        quick_edit_frame.set_default_content('Comments')
        quick_edit_frame.set_default_content('Stars')
        quick_edit_frame.set_default_content('Test result')
        quick_edit_frame.set_default_content('Must buy')
        quick_edit_frame.set_default_content('Installed folder')
        quick_edit_frame.set_default_content('Origin')
        quick_edit_frame.set_default_content('Alternative')

    def quick_edit_save_value(self, value: str, row: int = None, col: int = None) -> None:
        """
        Saves the cell content preview.
        :param value: The value to save.
        :param row: The row index of the cell.
        :param col: The column index of the cell.
        """
        if row is None or row >= len(self.model.df) or col is None:
            return
        try:
            self.model.df.iat[row, col] = value
            self.redraw()
            self.must_save = True
            log_debug(f'Save preview value {value} at row={row} col={col}')
        except IndexError:
            log_warning(f'Failed to save preview value {value} at row={row} col={col}')

    def get_image_url(self, row: int = None) -> str:
        """
        Returns the image URL of the selected row.
        :param row: The row index of the selected cell.
        :return: The image URL of the selected row.
        """
        if row is None:
            return ''
        try:
            return self.model.getValueAt(row, col=self.model.df.columns.get_loc('Image'))
        except IndexError:
            return ''

    def open_asset_url(self, url: str = None):
        """
        Opens the asset URL in a web browser.
        :param url: The URL to open.
        """
        if url is None:
            if self.edit_row_entries is None:
                return
            asset_url = self.edit_row_entries['Url'].get()
        else:
            asset_url = url
        log_info(f'calling open_asset_url={asset_url}')
        if asset_url is None or asset_url == '' or asset_url == gui_g.s.empty_cell:
            log_info('asset URL is empty for this asset')
            return
        webbrowser.open(asset_url)

    def reset_style(self) -> None:
        """
        Resets the table style. Usefull when style of the main ttk window has changed.
        """
        self.data.style.clear()
        self.redraw()
